import pytest
from sklearn import datasets
from py_muvr import permutation_test
from py_muvr.feature_selector import FeatureSelector
from py_muvr.utils import compute_t_student_p_value
from py_muvr.data_structures import (
    SelectedFeatures,
    ScoreCurve,
    FeatureSelectionResults,
)


@pytest.fixture
def fit_permutation_test(raw_results):
    fs = FeatureSelector(
        n_outer=5,
        metric="MISS",
        estimator="PLSC",
    )
    pt = permutation_test.PermutationTest(feature_selector=fs, n_permutations=4)

    pt.res = FeatureSelectionResults(
        raw_results,
        SelectedFeatures(min=[0, 1], max=[0, 1, 4, 3], mid=[0, 1, 4]),
        {"total": [ScoreCurve(n_features=[2, 3, 4, 5], scores=[4.0, 4.5, 4.0, 4.75])]},
    )
    pt.res_perm = [
        FeatureSelectionResults(
            raw_results,
            SelectedFeatures(min=[0], max=[0, 1, 2], mid=[0, 1]),
            {"total": [ScoreCurve(n_features=[1, 2, 3], scores=[4.0, 4.5, 4.5])]},
        ),
        FeatureSelectionResults(
            raw_results,
            SelectedFeatures(min=[0], max=[0, 1, 2], mid=[0, 1]),
            {"total": [ScoreCurve(n_features=[1, 2, 3], scores=[4.0, 4.5, 4.5])]},
        ),
        FeatureSelectionResults(
            raw_results,
            SelectedFeatures(min=[0], max=[0, 1, 2], mid=[0, 1]),
            {"total": [ScoreCurve(n_features=[1, 2, 3], scores=[4.0, 4.6, 4.9])]},
        ),
        FeatureSelectionResults(
            raw_results,
            SelectedFeatures(min=[0], max=[0, 1, 2], mid=[0, 1]),
            {"total": [ScoreCurve(n_features=[1, 2, 3], scores=[4.1, 4.6, 4.9])]},
        ),
    ]
    return pt


@pytest.fixture
def fit_feature_selector(raw_results):
    fs = FeatureSelector(
        n_outer=5,
        metric="MISS",
        estimator="PLSC",
    )
    fs.is_fit = True
    fs._raw_results = raw_results
    sel_feats = fs._post_processor.select_features(raw_results)
    fs._selected_features = sel_feats
    return fs


def test_permutation_test():
    fs = FeatureSelector(
        n_outer=8, n_repetitions=8, random_state=0, estimator="PLSC", metric="MISS"
    )
    assert permutation_test.PermutationTest(feature_selector=fs, n_permutations=10)


def test_copy_feature_selector(fit_feature_selector):
    pt = permutation_test.PermutationTest(fit_feature_selector, 2)
    fs_copy = pt._copy_feature_selector(fit_feature_selector)

    assert fit_feature_selector is not fs_copy
    assert fit_feature_selector.metric == fs_copy.metric
    assert fit_feature_selector.random_state == fs_copy.random_state
    assert fit_feature_selector.estimator == fs_copy.estimator
    assert fit_feature_selector.n_repetitions == fs_copy.n_repetitions
    assert fit_feature_selector.n_outer == fs_copy.n_outer
    assert fit_feature_selector.n_inner == fs_copy.n_inner
    assert fit_feature_selector.features_dropout_rate == fs_copy.features_dropout_rate
    assert fit_feature_selector.robust_minimum == fs_copy.robust_minimum


def test_fit(fit_feature_selector):
    n_permutations = 2
    pt = permutation_test.PermutationTest(
        feature_selector=fit_feature_selector, n_permutations=n_permutations
    )

    X, y = datasets.make_classification(n_features=5, n_samples=5, random_state=0)
    pt.fit(X, y)

    assert pt.res
    assert pt.res_perm
    assert len(pt.res_perm) == n_permutations


def test_compute_permutation_scores(fit_permutation_test):
    s_min, p_score_min = fit_permutation_test.compute_permutation_scores(model="min")
    s_mid, p_score_mid = fit_permutation_test.compute_permutation_scores(model="mid")
    s_max, p_score_max = fit_permutation_test.compute_permutation_scores(model="max")

    assert s_min == 4
    assert sorted(p_score_min) == [4.0, 4.0, 4.0, 4.1]
    assert s_mid == 4.5
    assert sorted(p_score_mid) == [4.5, 4.5, 4.6, 4.6]
    assert s_max == 4
    assert sorted(p_score_max) == [4.5, 4.5, 4.9, 4.9]


def test_compute_p_values(fit_permutation_test):
    min_p_val = fit_permutation_test.compute_p_values(model="min")
    mid_p_val = fit_permutation_test.compute_p_values(model="mid")
    max_p_val = fit_permutation_test.compute_p_values(model="max")

    assert min_p_val == compute_t_student_p_value(4, [4.0, 4.0, 4.0, 4.1])
    assert mid_p_val == compute_t_student_p_value(4.5, [4.5, 4.5, 4.6, 4.6])
    assert max_p_val == compute_t_student_p_value(4, [4.5, 4.5, 4.9, 4.9])


def test_rank_data(fit_permutation_test):
    rank_sample, ranks_population = fit_permutation_test._rank_data(10, [1, 2, 40])
    assert rank_sample == 3
    assert ranks_population[0] == 1
    assert ranks_population[1] == 2
    assert ranks_population[2] == 4



from typing import Union, Dict, List, Iterable

from py_muvr.data_structures import (
    RandomState,
    InputDataset,
    Split,
    TrainTestData,
)
from sklearn.model_selection import GroupShuffleSplit


class DataSplitter:
    """
    Class used to create the inner and outer splits for the double cross-validation and
    to slice the data using those splits and subsets of the features.

    Parameters
    ----------
    n_outer: int
        Number of outer splits to create
    n_inner:
        Number of inner splits to create
    input_data: InputDataset
        Object containing X, Y and groups attributes
    random_state: RandomState
        a random state instance to control reproducibility

    """

    def __init__(
        self,
        n_outer: int,
        n_inner: int,
        input_data: InputDataset,
        random_state: Union[int, RandomState],
    ):
        self.n_outer = n_outer
        self.n_inner = n_inner
        self.random_state = random_state
        self._splits = self._make_splits(input_data)

    def _make_splits(self, input_data: InputDataset) -> Dict[tuple, Split]:
        """Create inner and outer splits.

        Creates nested splits of the input data. The outer training datasets
        will be split again to provide validation and inner training datasets.
        """
        outer_splitter = self._make_random_splitter(self.n_outer)
        inner_splitter = self._make_random_splitter(self.n_inner)

        outer_splits = outer_splitter.split(
            input_data.X, input_data.y, input_data.groups
        )

        splits = {}
        for out_idx, (out_train, out_test) in enumerate(outer_splits):
            splits[(out_idx, None)] = Split(out_idx, out_train, out_test)
            X_train = input_data.X[out_train, :]
            y_train = input_data.y[out_train]
            groups_train = input_data.groups[out_train]
            inner_splits = inner_splitter.split(X_train, y_train, groups_train)
            for in_idx, (in_train, in_test) in enumerate(inner_splits):
                inner_split = Split(in_idx, out_train[in_train], out_train[in_test])
                splits[(out_idx, in_idx)] = inner_split
        return splits

    def _make_random_splitter(self, n_splits):
        test_size = 1 / n_splits
        return GroupShuffleSplit(
            n_splits, test_size=test_size, random_state=self.random_state
        )

    def iter_outer_splits(self) -> Iterable[Split]:
        """
        Iterates through the splits corresponding to the outer loops

        Returns
        -------
        Outer loop split used to slice the input dataset
        """
        for outer_idx in range(self.n_outer):
            yield self._splits[(outer_idx, None)]

    def iter_inner_splits(self, outer_split: Split) -> Iterable[Split]:
        """
        Given an outer split, iterates through the splits corresponding to the outer
        split's inner loops.

        Parameters
        ----------
        outer_split: Split
            Split from calling the iter_outer_splits method.

        Returns
        -------
        Split:
            Inner loop split used to slice the input dataset

        """
        outer_idx = outer_split.id
        for inner_idx in range(self.n_inner):
            yield self._splits[(outer_idx, inner_idx)]

    @staticmethod
    def split_data(
        input_data: InputDataset, split: Split, features: List[int] = None
    ) -> TrainTestData:
        """
        Splits the input dataset into train and test sets, optionally selecting a subset
        of features.

        Parameters
        ----------
        input_data: InputDataset
            Object containing X, Y and groups attributes
        split: Split
            Split from calling the iter splits methods
        features: List[int]
            List of features to select from the input data

        Returns
        -------
        TrainTestData:
            The sliced input data.
        """
        return TrainTestData(
            train_data=input_data[split.train_indices, features],
            test_data=input_data[split.test_indices, features],
        )



from typing import List

import numpy as np
from scipy.stats import mode
from sklearn.base import is_classifier

from py_muvr.feature_selector import FeatureSelector
from py_muvr.data_structures.data_types import NumpyArray
from py_muvr.models import Estimator


class ConsensusModel:
    def __init__(self, feature_selector: FeatureSelector, feature_set_label: str):
        """Implement consensus model based on the input feature selector.
        The members of the pool are the models trained on the outer loop test folds
        during `feature_selector.fit` execution. Only estimators relative to the input
        `feature_set_label` are considered.
        The user should also provide a correct `problem` argument, to identify
        the type of prediction the consensus model must perform. `problem`
        must be based on the estimator parameter that was used to build the
        `feature_selector`.

        Parameters
        ----------
        feature_selector : FeatureSelector
            The feature selector from which to take the outer loop models
        feature_set_label : str
            One of "min", "mid" and "max"

        Raises
        ------
        NotFitException
            If the feature selector was not fit yet
        ValueError
            If `feature_set` is not one of "min", "mid" or "max"

        Examples
        --------
        >>> fs = FeatureSelector(n_outer=6, metric="MISS", estimator="PLSC")
        >>> fs.fit(X, y)
        >>> cm = ConsensusModel(fs, "min", "classification")
        >>> y_pred = cm.predict(additional_X)
        """
        if feature_set_label not in {"min", "max", "mid"}:
            raise ValueError("'feature_set' must be one of 'min', 'mid' or 'max'")
        self._models = self._get_all_models(feature_selector, feature_set_label)
        self._feature_sets = self._get_all_feature_sets(
            feature_selector, feature_set_label
        )
        self.feature_set_label = feature_set_label

    @staticmethod
    def _get_all_feature_sets(selector: FeatureSelector, label: str) -> List[List[int]]:
        attr_name = label + "_eval"
        flat_results = [r for repetition in selector.raw_results for r in repetition]
        ranks = [getattr(result, attr_name).ranks for result in flat_results]
        feature_sets = [list(r.get_data().keys()) for r in ranks]
        return feature_sets

    @staticmethod
    def _get_all_models(selector: FeatureSelector, label: str) -> List[Estimator]:
        flat_results = [r for repetition in selector.raw_results for r in repetition]
        attr_name = label + "_eval"
        return [getattr(result, attr_name).model for result in flat_results]

    def _ensemble_predict(self, X: NumpyArray) -> NumpyArray:
        """Perform a prediction for the input `X` for each one of the outer loop models
        of the input feature selector.

        Parameters
        ----------
        X : NumpyArray
            Predictor variables as numpy array

        Returns
        -------
        NumpyArray
            Predicted response vectors, as matrix with as many rows as models

        Raises
        ------
        ValueError
            if input X number of columns doesn't match the feature_selector number
            of features
        """
        y_preds = []
        for feats, model in zip(self._feature_sets, self._models):
            pred_X = X[:, feats]
            y_pred = model.predict(pred_X)
            y_preds.append(y_pred)
        y_preds = np.vstack(y_preds)  # n-rows = n-models
        return y_preds

    def predict(self, X: NumpyArray) -> NumpyArray:
        """Perform ensemble prediction aggregating single predictions from the
        feature_selector models. If the problem is a classification the aggregation is
        the mode of the predicted classes, otherwise is the mean of the predicted
        response vectors.

        Parameters
        ----------
        X : NumpyArray
            Predictor variables as numpy array

        Returns
        -------
        NumpyArray
            Predicted response vector
        """
        y_preds = self._ensemble_predict(X)
        if is_classifier(self._models[0]):
            return mode(y_preds, axis=0)[0].ravel()
        return y_preds.mean(axis=0)



from typing import Union, List
from scipy.stats import rankdata
from py_muvr.data_structures import (
    MetricFunction,
    RandomState,
    NumpyArray,
    FeatureEvaluationResults,
    FeatureRanks,
    TrainTestData,
)
from py_muvr.models.metrics import make_metric
from py_muvr.models import make_estimator, Estimator


class FeatureEvaluator:
    """
    This class is used to evaluate a set of features given an estimator, a scoring
    metric and a random state.

    Parameters
    ----------
    estimator: Estimator
        Model used to evaluate feature sets
    metric: Union[str, MetricFunction]
        Metric used to measure the performance of the model using the feature set.
    random_state: Union[int, RandomState]
        A random state instance to control reproducibility
    """

    def __init__(
        self,
        estimator: Estimator,
        metric: Union[str, MetricFunction],
        random_state: Union[int, RandomState],
    ):
        self._estimator = make_estimator(estimator, random_state)
        self._metric = make_metric(metric)
        self._n_initial_features = 0

    def set_n_initial_features(self, n_initial_features: int):
        """
        Saves the initial number of features to be used as a reference to compute the
        average of all feature ranks from different feature subsets.

        Parameters
        ----------
        n_initial_features: int
            The initial number of features to use as a reference for feature ranks.

        """
        self._n_initial_features = n_initial_features

    def evaluate_features(
        self, evaluation_data: TrainTestData, features: List[int]
    ) -> FeatureEvaluationResults:
        """
        This method evaluates a feature set on an input data. This is done by training
        the estimator on a train set, predicting for a validation set and computing the
        score on the validation set based on the chosen metric.

        Parameters
        ----------
        evaluation_data: TrainTestData
            Dataset object containing train/test X and y
        features:
            Feature set being used on the evaluation data

        Returns
        -------
        FeatureEvaluationResults:
            The results of the evaluation. It consists of:
            - features: the feature set
            - ranks: the ranking of the features
            - model: the estimator used on evaluation

        """
        if self._n_initial_features == 0:
            raise ValueError("Call set_n_initial_features first")

        X_train = evaluation_data.train_data.X
        y_train = evaluation_data.train_data.y
        X_test = evaluation_data.test_data.X
        y_test = evaluation_data.test_data.y

        estimator = self._estimator.clone().fit(X_train, y_train)
        y_pred = estimator.predict(X_test)

        score = -self._metric(y_test, y_pred)
        ranks = self._get_feature_ranks(estimator, features)
        return FeatureEvaluationResults(test_score=score, ranks=ranks, model=estimator)

    def _get_feature_ranks(
        self, estimator: Estimator, features: Union[List[int], NumpyArray]
    ) -> FeatureRanks:
        feature_importances = estimator.feature_importances
        ranks = rankdata(-feature_importances)
        return FeatureRanks(
            features=features, ranks=ranks, n_feats=self._n_initial_features
        )



import pytest

import collections
from py_muvr.post_processor import PostProcessor
from py_muvr.feature_evaluator import FeatureEvaluationResults
from py_muvr.data_structures import FeatureRanks, ScoreCurve, OuterLoopResults
from py_muvr.models import Estimator


@pytest.fixture
def outer_loop_results():
    return [
        OuterLoopResults(
            n_features_to_score_map={5: 100, 4: 5, 3: 4, 2: 5, 1: 100},
            min_eval=FeatureEvaluationResults(
                test_score=0,
                ranks=FeatureRanks(features=[1, 2], ranks=[1, 2], n_feats=5),
                model="model",
            ),
            max_eval=FeatureEvaluationResults(
                test_score=0,
                ranks=FeatureRanks(
                    features=[1, 2, 3, 4], ranks=[2, 1, 3, 4], n_feats=5
                ),
                model="model",
            ),
            mid_eval=FeatureEvaluationResults(
                test_score=0,
                ranks=FeatureRanks(features=[1, 2, 3], ranks=[1, 2, 3], n_feats=5),
                model="model",
            ),
        ),
        OuterLoopResults(
            n_features_to_score_map={5: 300, 4: 6, 3: 4, 2: 7, 1: 250},
            min_eval=FeatureEvaluationResults(
                test_score=0,
                ranks=FeatureRanks(features=[1, 2], ranks=[1.5, 1.5], n_feats=5),
                model="model",
            ),
            max_eval=FeatureEvaluationResults(
                test_score=0,
                ranks=FeatureRanks(
                    features=[0, 1, 2, 3], ranks=[1, 2, 3, 4], n_feats=5
                ),
                model="model",
            ),
            mid_eval=FeatureEvaluationResults(
                test_score=0,
                ranks=FeatureRanks(features=[0, 1, 2], ranks=[3, 1, 2], n_feats=5),
                model="model",
            ),
        ),
    ]


@pytest.fixture
def outer_loop_results2():
    return [
        OuterLoopResults(
            n_features_to_score_map={5: 150, 4: 4, 3: 4, 2: 5, 1: 120},
            min_eval=FeatureEvaluationResults(
                test_score=0,
                ranks=FeatureRanks(features=[2, 3], ranks=[1, 2], n_feats=5),
                model="model",
            ),
            max_eval=FeatureEvaluationResults(
                test_score=0,
                ranks=FeatureRanks(
                    features=[0, 1, 2, 3], ranks=[3, 1, 2, 4], n_feats=5
                ),
                model="model",
            ),
            mid_eval=FeatureEvaluationResults(
                test_score=0,
                ranks=FeatureRanks(features=[0, 1, 2], ranks=[3, 1, 2], n_feats=5),
                model="model",
            ),
        ),
        OuterLoopResults(
            n_features_to_score_map={5: 200, 4: 7, 3: 1, 2: 6, 1: 220},
            min_eval=FeatureEvaluationResults(
                test_score=0,
                ranks=FeatureRanks(features=[0, 1], ranks=[1, 2], n_feats=5),
                model="model",
            ),
            max_eval=FeatureEvaluationResults(
                test_score=0,
                ranks=FeatureRanks(
                    features=[0, 1, 2, 4], ranks=[4, 3, 1, 2], n_feats=5
                ),
                model="model",
            ),
            mid_eval=FeatureEvaluationResults(
                test_score=0,
                ranks=FeatureRanks(features=[1, 2, 3], ranks=[3, 1, 2], n_feats=5),
                model="model",
            ),
        ),
    ]


@pytest.fixture
def repetitions(outer_loop_results, outer_loop_results2):
    return [outer_loop_results, outer_loop_results2]


def test_post_processor():
    pp = PostProcessor(robust_minimum=0.05)
    assert pp


def test_select_features(repetitions):
    pp = PostProcessor(robust_minimum=0.05)
    selected_feats = pp.select_features(repetitions)
    # the scores have a deep minimum at 2, 3 and 4 features
    assert selected_feats["min"]
    assert selected_feats["mid"]
    assert selected_feats["max"]
    assert sorted(selected_feats["min"]) == [1, 2]
    assert sorted(selected_feats["mid"]) == [1, 2, 3]
    assert sorted(selected_feats["max"]) == [0, 1, 2, 3]


def test_compute_n_features(repetitions):
    pp = PostProcessor(robust_minimum=0.05)
    n_feats = pp._compute_n_features(repetitions)
    # the scores have a deep minimum at 2, 3 and 4 features
    assert len(n_feats) == 3
    min_feats, mid_feats, max_feats = n_feats
    assert min_feats == 2
    assert mid_feats == 3
    assert max_feats == 4


def test_get_repetition_avg_scores(repetitions):
    pp = PostProcessor(robust_minimum=0.05)
    avg_scores = pp._get_repetition_avg_scores(repetitions)
    assert len(avg_scores) == 2
    assert avg_scores[0][5] == 200
    assert avg_scores[0][4] == 5.5
    assert avg_scores[0][3] == 4
    assert avg_scores[0][2] == 6
    assert avg_scores[0][1] == 175


def test_get_validation_curves(repetitions):
    pp = PostProcessor(robust_minimum=0.05)
    curves = pp.get_validation_curves(repetitions)
    assert len(curves) == 3
    assert len(curves["outer_loops"]) == 4
    assert len(curves["repetitions"]) == 2
    assert len(curves["total"]) == 1
    assert isinstance(curves["total"][0], ScoreCurve)
    assert sorted(curves["outer_loops"][0].n_features) == [1, 2, 3, 4, 5]
    assert sorted(curves["repetitions"][0].n_features) == [1, 2, 3, 4, 5]
    assert list(curves["repetitions"][0].scores) == [175, 6, 4, 5.5, 200]


def test_process_feature_elim_results(rfe_raw_results):
    pp = PostProcessor(0.05)

    processed = pp.process_feature_elim_results(rfe_raw_results)

    assert processed.best_features
    assert processed.n_features_to_score_map


def test_compute_score_curve(rfe_raw_results):
    pp = PostProcessor(0.05)

    avg_scores = pp._compute_score_curve(rfe_raw_results)

    assert len(avg_scores) == 3
    assert 2 in avg_scores
    assert 3 in avg_scores
    assert 4 in avg_scores
    assert avg_scores[4] < avg_scores[3]


def test_select_best_features(rfe_raw_results):
    pp = PostProcessor(1)
    avg_scores = pp._compute_score_curve(rfe_raw_results)

    selected_feats = pp._select_best_outer_features(rfe_raw_results, avg_scores)

    assert sorted(selected_feats["min"]) == [2, 4]
    assert sorted(selected_feats["mid"]) == [2, 3, 4]
    assert sorted(selected_feats["max"]) == [1, 2, 3, 4]


def test_make_average_ranks_dataframe(fs_results):
    pp = PostProcessor(1)
    n_feats = 5
    feature_names = "a, b, c, d, e".split(", ")

    ranks_df = pp.make_average_ranks_df(fs_results, n_feats, feature_names)

    assert ranks_df.ndim
    assert len(ranks_df) == n_feats
    assert set(ranks_df.index) == set(feature_names)


def test_exclude_unused_features(fs_results):
    pp = PostProcessor(1)
    n_feats = 5
    unused_feats = 10

    reduced_df = pp.make_average_ranks_df(
        fs_results, unused_feats, exclude_unused_features=True
    )
    full_df = pp.make_average_ranks_df(
        fs_results, unused_feats, exclude_unused_features=False
    )

    assert reduced_df.ndim
    assert full_df.ndim
    assert len(reduced_df) == n_feats
    assert len(full_df) == unused_feats


def test_get_feature_ranks(raw_results):
    pp = PostProcessor(1)

    min_ranks = pp._get_feature_ranks(raw_results, "min")
    max_ranks = pp._get_feature_ranks(raw_results, "max")

    assert len(min_ranks) == len(max_ranks)
    assert isinstance(min_ranks[0], dict)



import os
import pathlib
import re
import string
import numpy as np
import pytest
import pandas as pd
from sklearn.datasets import make_classification

PYTHON_SNIPPET_REGEX = re.compile(r"```python([^`]+)```")


@pytest.fixture
def data():
    n_feats = 10
    n_samples = 10
    feature_names = [string.ascii_lowercase[i] for i in range(n_feats)] + ["target"]
    X, y = make_classification(n_feats, n_samples, random_state=0)
    data = pd.DataFrame(
        data=np.hstack((X, y.reshape((n_samples, 1)))), columns=feature_names
    )
    return data


@pytest.fixture
def readme_string():
    readme_dir = pathlib.Path(__file__).parent.parent.absolute()
    with open(os.path.join(readme_dir, "README.md")) as f:
        readme = f.read()
    return readme


@pytest.mark.skip(os.getenv("CI", "False").title())
@pytest.mark.slow
def test_readme_code_examples(data, readme_string):
    """this tests checks that every (almost) python snippet in the README
    would run sequentially without raising errors"""
    python_snippets = PYTHON_SNIPPET_REGEX.findall(readme_string)
    for snippet in python_snippets:
        if "read_csv" in snippet:  # the file is an unexisting example
            continue
        exec(snippet.strip())



from pathlib import Path

import numpy as np
import pandas as pd
import pytest

from py_muvr import FeatureSelector
from py_muvr.data_structures import (
    FeatureEvaluationResults,
    FeatureRanks,
    InputDataset,
    OuterLoopResults,
)

ASSETS_DIR = Path(__file__).parent / "assets"


@pytest.fixture(scope="session")
def raw_results():
    return [
        [
            OuterLoopResults(
                min_eval=FeatureEvaluationResults(
                    test_score=4,
                    model="model",
                    ranks=FeatureRanks(features=[0, 1], ranks=[1, 2], n_feats=10),
                ),
                max_eval=FeatureEvaluationResults(
                    test_score=5,
                    model="model",
                    ranks=FeatureRanks(
                        features=[0, 1, 2, 3], ranks=[1, 2, 4, 3], n_feats=10
                    ),
                ),
                mid_eval=FeatureEvaluationResults(
                    test_score=5,
                    model="model",
                    ranks=FeatureRanks(features=[0, 1, 3], ranks=[1, 2, 3], n_feats=10),
                ),
                n_features_to_score_map={5: 4, 4: 3, 3: 3, 2: 3},
            ),
            OuterLoopResults(
                min_eval=FeatureEvaluationResults(
                    test_score=3,
                    model="model",
                    ranks=FeatureRanks(
                        features=[0, 1, 4, 3], ranks=[1, 2, 3, 4], n_feats=10
                    ),
                ),
                max_eval=FeatureEvaluationResults(
                    test_score=3,
                    model="model",
                    ranks=FeatureRanks(
                        features=[0, 1, 4, 3], ranks=[1, 2, 3, 4], n_feats=10
                    ),
                ),
                mid_eval=FeatureEvaluationResults(
                    test_score=2,
                    model="model",
                    ranks=FeatureRanks(
                        features=[0, 1, 4, 3], ranks=[1, 2, 3, 4], n_feats=10
                    ),
                ),
                n_features_to_score_map={5: 5, 4: 4, 3: 5, 2: 5},
            ),
        ],
        [
            OuterLoopResults(
                min_eval=FeatureEvaluationResults(
                    test_score=4,
                    model="model",
                    ranks=FeatureRanks(features=[0, 1], ranks=[1, 2], n_feats=10),
                ),
                max_eval=FeatureEvaluationResults(
                    test_score=5,
                    model="model",
                    ranks=FeatureRanks(
                        features=[0, 1, 4, 2], ranks=[1, 2, 3, 4], n_feats=10
                    ),
                ),
                mid_eval=FeatureEvaluationResults(
                    test_score=5,
                    model="model",
                    ranks=FeatureRanks(features=[0, 1, 4], ranks=[2, 1, 3], n_feats=10),
                ),
                n_features_to_score_map={5: 5, 4: 3, 3: 5, 2: 3},
            ),
            OuterLoopResults(
                min_eval=FeatureEvaluationResults(
                    test_score=2,
                    model="model",
                    ranks=FeatureRanks(features=[0, 1], ranks=[1, 2], n_feats=10),
                ),
                max_eval=FeatureEvaluationResults(
                    test_score=2,
                    model="model",
                    ranks=FeatureRanks(
                        features=[0, 1, 2, 3, 4], ranks=[1, 2, 5, 4, 3], n_feats=10
                    ),
                ),
                mid_eval=FeatureEvaluationResults(
                    test_score=2,
                    model="model",
                    ranks=FeatureRanks(features=[0, 1, 4], ranks=[1, 2, 3], n_feats=10),
                ),
                n_features_to_score_map={5: 5, 4: 6, 3: 5, 2: 5},
            ),
        ],
    ]


@pytest.fixture
def inner_loop_results():
    return [
        FeatureEvaluationResults(
            ranks=FeatureRanks(features=[1, 2, 3, 4], ranks=[3, 2, 1, 4]),
            test_score=0.2,
            model="estimator",
        ),
        FeatureEvaluationResults(
            ranks=FeatureRanks(features=[1, 2, 3, 4], ranks=[1.5, 1.5, 3, 4]),
            test_score=0.2,
            model="estimator",
        ),
    ]


@pytest.fixture
def inner_loop_results_2():
    return [
        FeatureEvaluationResults(
            ranks=FeatureRanks(features=[2, 3, 4], ranks=[3, 2, 1]),
            test_score=0.1,
            model="model",
        ),
        FeatureEvaluationResults(
            ranks=FeatureRanks(features=[2, 3, 4], ranks=[1.5, 1.5, 3]),
            test_score=0.5,
            model="model",
        ),
    ]


@pytest.fixture
def inner_loop_results_3():
    return [
        FeatureEvaluationResults(
            ranks=FeatureRanks(features=[2, 4], ranks=[3, 2, 1]),
            test_score=0.3,
            model="model",
        ),
        FeatureEvaluationResults(
            ranks=FeatureRanks(features=[2, 4], ranks=[1.5, 1.5, 3]),
            test_score=0.25,
            model="model",
        ),
    ]


@pytest.fixture
def rfe_raw_results(inner_loop_results, inner_loop_results_2, inner_loop_results_3):
    return {
        (1, 2, 3, 4): inner_loop_results,
        (2, 3, 4): inner_loop_results_2,
        (2, 4): inner_loop_results_3,
    }


@pytest.fixture
def dataset():
    X = np.random.rand(12, 12)
    y = np.random.choice([0, 1], 12)
    return InputDataset(X=X, y=y, groups=np.arange(12))


@pytest.fixture(scope="session")
def mosquito():
    df = pd.read_csv(ASSETS_DIR / "mosquito.csv", index_col=0)
    df = df.sample(frac=1)
    X = df.drop(columns=["Yotu"]).values
    y = df.Yotu.values
    groups = df.index
    return InputDataset(X=X, y=y, groups=groups)


@pytest.fixture(scope="session")
def freelive():
    df = pd.read_csv(ASSETS_DIR / "freelive.csv", index_col=0)
    X = df.drop(columns=["YR"]).values
    y = df.YR.values
    groups = df.index
    return InputDataset(X=X, y=y, groups=groups)


@pytest.fixture(scope="session")
def fs_results(raw_results):
    fs = FeatureSelector(n_outer=3, metric="MISS", estimator="RFC")
    fs._raw_results = raw_results
    fs.is_fit = True
    fs._selected_features = fs._post_processor.select_features(raw_results)
    fs._n_features = 5
    fs_results = fs.get_feature_selection_results(["A", "B", "C", "D", "E"])
    return fs_results



from concurrent.futures import Future, Executor


class SyncExecutor(Executor):
    """An executor which immediately invokes all submitted callables."""

    def submit(self, fn, *args, **kwargs):  # pylint: disable=arguments-differ
        """Immediately invokes `fn(*args, **kwargs)` and returns a future
        with the result (or exception)."""
        future = Future()
        try:
            result = fn(*args, **kwargs)
            future.set_result(result)
        except Exception as e:  # pylint: disable=broad-except
            future.set_exception(e)
        return future



import pytest
import numpy as np
from py_muvr.models import make_estimator
from sklearn.ensemble import RandomForestRegressor
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import Normalizer
from sklearn.svm import SVC, SVR
from sklearn.linear_model import LogisticRegression, LinearRegression

models = [
    RandomForestRegressor(10),
    SVR(kernel="linear"),
    SVC(kernel="linear", random_state=1),
    Pipeline([("norm", Normalizer()), ("model", SVC(kernel="linear", random_state=1))]),
    Pipeline([("norm", Normalizer()), ("model", LinearRegression())]),
    LinearRegression(),
    LogisticRegression(),
]


@pytest.mark.parametrize("model", models)
def test_get_feature_importances(model, dataset):
    estimator = make_estimator(model, np.random.RandomState(0))
    estimator.fit(dataset.X, dataset.y)
    assert len(estimator.feature_importances) == dataset.X.shape[1]



import logging
from typing import Union
import numpy as np
from sklearn.cross_decomposition import PLSRegression
from sklearn.preprocessing import OneHotEncoder
from py_muvr.data_structures.data_types import NumpyArray


log = logging.getLogger(__name__)


def get_vip(model: PLSRegression) -> NumpyArray:
    # Calculate VIP
    # https://github.com/scikit-learn/scikit-learn/pull/13492
    # https://www.ncbi.nlm.nih.gov/pmc/articles/PMC3883573/#S6title
    T = model.x_scores_
    W = model.x_weights_
    Q = model.y_loadings_
    w0, w1 = W.shape
    s = np.sum(T ** 2, axis=0) * np.sum(Q ** 2, axis=0)
    s_sum = np.sum(s, axis=0)
    w_norm = np.array([(W[:, i] / np.linalg.norm(W[:, i])) for i in range(w1)])
    vip = np.sqrt(w0 * np.sum(s * w_norm.T ** 2, axis=1) / s_sum)
    return vip


class PLSClassifier(PLSRegression):
    """This class extends the scikit-learn PLS regression to
    multi-class classification. To do so, it converts the classification problem to
    a multi dimensional regression. Output classes are one-hot encoded and used
    to train a PLS regressor. Predictions are given by casting the output
    of the regression to a binary matrix where the 1s correspond to the maximum
    of each row. Finally the matrix is converted to classes using the inverse one hot
    transformation.

    It also provides `feature_importances_` as VIP and prevent fit failures
    if the number of components is bigger than the number of features

    The signature is analogous to sklearn.cross_decomposition.PLSRegression:

    Parameters
    ----------
    n_components: int, optional
        Number of components to keep, by default 2.
    scale: bool, optional
        whether to scale the data, by default True.
    max_iter: int, optional
        the maximum number of iterations of the NIPALS inner loop, by default 500.
    tol: float, optional
        tolerance used in the iterative algorithm, by default 1e-06.
    copy: bool, optional
        Whether the deflation should be done on a copy. Let the default value to True
        unless you don’t care about side effects, by default True.

    """

    encoder = None
    feature_importances_ = None
    _estimator_type = "classifier"

    def fit(self, X: NumpyArray, Y: NumpyArray):
        if self.n_components > X.shape[1]:
            reduce_pls_components(self, X.shape[1])
        self.encoder = OneHotEncoder().fit(Y.reshape(-1, 1))
        encoded_y = self.encoder.transform(Y.reshape(-1, 1)).toarray()
        super().fit(X, encoded_y)
        self.feature_importances_ = get_vip(self)
        return self

    def predict(self, X: NumpyArray, copy: bool = True):
        y_pred = super().predict(X, copy=copy)
        y_pred = np.equal(y_pred.T, y_pred.max(axis=1)).T.astype(float)
        return self.encoder.inverse_transform(y_pred).ravel()


class PLSRegressor(PLSRegression):
    """This class extends the scikit-learn PLS regression.
    It provides `feature_importances_` as VIP and prevent fit failures
    if the number of components is bigger than the number of features

    The signature is analogous to sklearn.cross_decomposition.PLSRegression:

    Parameters
    ----------
    n_components: int, optional
        Number of components to keep, by default 2.
    scale: bool, optional
        whether to scale the data, by default True.
    max_iter: int, optional
        the maximum number of iterations of the NIPALS inner loop, by default 500.
    tol: float, optional
        tolerance used in the iterative algorithm, by default 1e-06.
    copy: bool, optional
        Whether the deflation should be done on a copy. Let the default value to True
        unless you don’t care about side effects, by default True.

    """

    feature_importances_ = None

    def fit(self, X: NumpyArray, Y: NumpyArray):
        if self.n_components > X.shape[1]:
            reduce_pls_components(self, X.shape[1])
        super().fit(X, Y)
        self.feature_importances_ = get_vip(self)
        return self

    def predict(self, X: NumpyArray, copy: bool = True):
        return super().predict(X, copy=copy).ravel()


def reduce_pls_components(pls: Union[PLSRegressor, PLSClassifier], n_components: int):
    # there might be occasions in which the loops try to fit on a matrix
    # with n_features, a PLS with n_components > n_features. The recursive
    # feature elimination where we stop at n_features = 1 is a clear case
    # of this
    log.debug("Lowering PLS n_components to %s during fit", n_components)
    pls.set_params(n_components=n_components)



import pytest
import logging
import numpy as np
from py_muvr import utils
from py_muvr.data_structures import FeatureRanks


@pytest.fixture
def scores():
    return [{5: 5, 4: 4, 3: 4, 2: 4}, {5: 10, 4: 9, 3: 10, 2: 11}]


def test_average_scores(scores):
    avg_score = utils.average_scores(scores)
    assert avg_score
    assert avg_score[5] == 7.5
    assert avg_score[4] == 6.5
    assert avg_score[3] == 7
    assert avg_score[2] == 7.5


def test_normalize_score():
    avg_score = {1: 11, 2: 6, 3: 1, 4: 6, 5: 11}
    norm_score = utils.normalize_score(avg_score)
    assert norm_score
    assert norm_score[1] == 1
    assert norm_score[3] == 0
    assert norm_score[2] == 0.5


def test_average_ranks():
    features1 = [1, 2, 3]
    features2 = [0, 1, 4, 5]
    ranks1 = FeatureRanks(features=features1, ranks=[1, 2, 3], n_feats=10)
    ranks2 = FeatureRanks(features=features2, ranks=[1, 2, 3, 4], n_feats=10)
    avg_ranks = utils.average_ranks([ranks1, ranks2])
    assert isinstance(avg_ranks, FeatureRanks)
    assert avg_ranks[0] == 5.5
    assert avg_ranks[1] == 1.5
    assert avg_ranks[2] == 6


@pytest.mark.parametrize(
    "n, best",
    [
        (1, [5]),
        (2, [5, 0]),
        (3, [5, 0, 1]),
        (4, [5, 0, 1, 4]),
    ],
)
def test_get_best_ranks(n, best):
    ranks = FeatureRanks(features=[5, 0, 1, 4], ranks=[1, 2, 3, 4], n_feats=10)
    best_feats = utils.get_best_n_features(ranks, n)
    assert sorted(best_feats) == sorted(best)


def test_compute_t_student_p_value():
    """Using p-values tables for t-student with 17 dof"""
    dof = 17
    n = dof + 1
    random_state = np.random.RandomState(9)
    sample_005 = -1.74
    sample_01 = -1.333
    p_values_005 = []
    p_values_01 = []
    for i in range(1000):
        population = random_state.normal(size=n)
        p_value = utils.compute_t_student_p_value(sample_005, population)
        p_values_005.append(p_value)
        p_value = utils.compute_t_student_p_value(sample_01, population)
        p_values_01.append(p_value)
    assert abs(np.mean(p_values_005) - 0.05) < 0.01
    assert abs(np.mean(p_values_01) - 0.1) < 0.01


def test_mute_loggers():
    @utils.mute_loggers(loggers=["py_muvr.feature_selector"])
    def test_function(logger_name):
        return logging.getLogger(logger_name).getEffectiveLevel()

    @utils.mute_loggers(loggers=["py_muvr.feature_selector"])
    def test_bad_function(logger_name):
        raise RuntimeError("exception")

    assert test_function("py_muvr.models.pls") == logging.INFO
    assert test_function("py_muvr.feature_selector") == logging.WARN
    assert logging.getLogger("py_muvr.models.pls").getEffectiveLevel() == logging.INFO
    assert (
        logging.getLogger("py_muvr.feature_selector").getEffectiveLevel()
        == logging.INFO
    )
    with pytest.raises(RuntimeError):
        test_bad_function("py_muvr.feature_selector")
    assert (
        logging.getLogger("py_muvr.feature_selector").getEffectiveLevel()
        == logging.INFO
    )



from typing import Union
from py_muvr.data_structures import MetricFunction, NumpyArray
from py_muvr.models.sklearn_metrics import SKLEARN_METRICS


def make_metric(metric: Union[str, MetricFunction]) -> MetricFunction:
    """Build metric function using the input `metric`. If a metric is a string
    then is interpreted as a scikit-learn metric score, such as "accuracy".
    Else, if should be a callable accepting two input arrays."""
    if isinstance(metric, str):
        return _make_metric_from_string(metric)
    if hasattr(metric, "__call__"):
        return metric
    raise ValueError("Input metric is not valid")


def _make_metric_from_string(metric_string: str) -> MetricFunction:
    if metric_string == "MISS":
        return miss_score
    if metric_string in SKLEARN_METRICS:
        return SKLEARN_METRICS[metric_string]
    raise ValueError("Input metric is not a valid string")


def miss_score(y_true: NumpyArray, y_pred: NumpyArray):
    """MISS score: number of wrong classifications preceded by - so that the higher
    this score the better the model"""
    return -(y_true != y_pred).sum()



from abc import ABC, abstractmethod
from py_muvr.data_structures.data_types import RandomState


class Estimator(ABC):
    @property
    @abstractmethod
    def feature_importances(self):
        pass

    @property
    @abstractmethod
    def _estimator_type(self):
        pass

    @abstractmethod
    def set_random_state(self, random_state: RandomState):
        pass

    @abstractmethod
    def clone(self):
        pass

    @abstractmethod
    def fit(self, X, y):
        pass

    @abstractmethod
    def predict(self, X):
        pass



import pytest
import numpy as np
from sklearn.ensemble import RandomForestRegressor
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import Normalizer
from sklearn.svm import SVC, SVR
from sklearn.linear_model import LogisticRegression, LinearRegression
from sklearn.exceptions import NotFittedError
from py_muvr.models.estimator import Estimator
from py_muvr.models import make_estimator

models = [
    "RFC",
    RandomForestRegressor(10),
    SVR(kernel="linear"),
    "XGBC",
    "PLSC",
    "PLSR",
    SVC(kernel="linear", random_state=1),
    Pipeline([("norm", Normalizer()), ("model", SVC(kernel="linear", random_state=1))]),
    LinearRegression(),
    LogisticRegression(),
]


@pytest.mark.parametrize("est", models)
def test_make_estimator(est):
    estimator = make_estimator(est, 0)
    assert estimator
    assert isinstance(estimator, Estimator)


@pytest.mark.parametrize("est", ["yooo", SVC])
def test_make_estimator_errors(est):
    with pytest.raises(ValueError):
        estimator = make_estimator(est, 0)


@pytest.mark.parametrize("est", models)
def test_estimator_fit(est, dataset):
    estimator = make_estimator(est, 0)
    fit_estimator = estimator.fit(dataset.X, dataset.y)
    assert estimator is fit_estimator


@pytest.mark.parametrize(
    "est",
    models,
)
def test_estimator_predict(est, dataset):
    estimator = make_estimator(est, 0)
    with pytest.raises(NotFittedError):
        _ = estimator.predict(dataset.X)


@pytest.mark.parametrize(
    "est",
    models,
)
def test_estimator_predict(est, dataset):
    estimator = make_estimator(est, 0).fit(dataset.X, dataset.y)
    y = estimator.predict(dataset.X)
    assert y.size == dataset.y.size
    assert y.shape == dataset.y.shape


@pytest.mark.parametrize("est", models)
def test_estimator_clone(est, dataset):
    estimator = make_estimator(est, 0)
    estimator = estimator.fit(dataset.X, dataset.y)
    cloned = estimator.clone()
    assert cloned is not estimator
    with pytest.raises(NotFittedError):
        y_pred = cloned.predict(dataset.X)


@pytest.mark.parametrize("est", models)
def test_train_model(est, dataset):
    estimator = make_estimator(est, 0)
    trained_estimator = estimator.clone().fit(dataset.X, dataset.y)
    assert estimator is not trained_estimator


@pytest.mark.parametrize("estimator", models)
def test_get_feature_importancres(estimator, dataset):
    estimator = make_estimator(estimator, 0)
    estimator.fit(dataset.X, dataset.y)
    feature_importances = estimator.feature_importances
    assert any(feature_importances)
    assert all(feature_importances >= 0)
    assert len(feature_importances) == 12



from unittest.mock import Mock
import pytest
from concurrent.futures import ProcessPoolExecutor

import numpy as np
import pandas as pd
from sklearn.linear_model import LinearRegression
from loky import get_reusable_executor
from dask.distributed import Client

from py_muvr.data_structures import (
    InputDataset,
    SelectedFeatures,
    FeatureSelectionResults,
    ScoreCurve,
)
from py_muvr.data_splitter import DataSplitter
from py_muvr.feature_selector import FeatureSelector


@pytest.fixture()
def fs():
    lr = LinearRegression()
    fs = FeatureSelector(
        n_outer=4,
        n_repetitions=4,
        random_state=0,
        estimator=lr,
        metric="MISS",
        features_dropout_rate=0.1,
    )
    return fs


def test_feature_selector():
    fs = FeatureSelector(
        n_outer=8,
        metric="MISS",
        estimator="RFC",
        features_dropout_rate=0.05,
        robust_minimum=0.05,
        n_repetitions=8,
        random_state=0,
    )
    assert fs
    assert fs.n_inner == 7


def test_fit(fs):
    X = np.random.rand(10, 10)
    y = np.array([np.random.choice([0, 1]) for _ in range(10)])
    fitted_fs = fs.fit(X, y)
    assert fitted_fs is fs
    assert fs._selected_features
    assert fs.is_fit


def test_get_groups(fs):
    predefined_group = [1, 2, 3]
    generated_groups = fs._get_groups(None, 5)

    assert all(generated_groups == [0, 1, 2, 3, 4])
    assert fs._get_groups(predefined_group, 5) == predefined_group


def test_run_outer_loop(fs):
    x = np.array([[2, 3, 4], [2, 3, 4]])
    input_data = InputDataset(x, "y", "groups")
    data_splitter = Mock(DataSplitter)
    data_splitter.split_data = Mock(spec=data_splitter.split_data, return_value="split")

    data_splitter.iter_inner_splits = Mock(
        data_splitter.iter_inner_splits, return_value=[1, 2]
    )

    fs._remove_features = Mock(return_value=[])
    fs._create_outer_loop_results = Mock(
        spec=fs._create_outer_loop_results, return_value="outer_loop_res"
    )
    fs._feature_evaluator.evaluate_features = Mock(
        spec=fs._feature_evaluator.evaluate_features, return_value="res"
    )
    fs._post_processor.process_feature_elim_results = Mock(
        fs._post_processor.process_feature_elim_results,
        return_value="processed_results",
    )
    inner_results = ["res", "res"]
    features = [0, 1, 2]
    raw_results = {tuple(features): inner_results}

    olr = fs._run_outer_loop(input_data, "outer_split", data_splitter)

    assert olr == "outer_loop_res"
    data_splitter.iter_inner_splits.assert_called_with("outer_split")
    data_splitter.split_data.assert_called_with(input_data, 2, features)
    fs._remove_features.assert_called_once_with(features, inner_results)
    fs._feature_evaluator.evaluate_features.assert_called_with("split", features)
    fs._create_outer_loop_results.assert_called_with(
        raw_results, input_data, "outer_split", data_splitter
    )


def test_remove_features(fs, inner_loop_results):
    features = [1, 2, 3, 4]
    fs.keep_fraction = 0.75

    features = fs._remove_features(features, inner_loop_results)

    assert len(features) == 3
    assert features == [2, 3, 1]


def test_select_n_best(fs, inner_loop_results):
    keep = 2
    n_best = fs._select_n_best(inner_loop_results, keep)

    assert n_best == [2, 3]


def test_create_outer_loop_results():
    pass  # TODO


def test_evaluate_min_mid_and_max_features(fs, dataset, rfe_raw_results):
    best_features = SelectedFeatures([1, 2], [1, 2, 3, 4], [1, 2, 3])
    fs._feature_evaluator.evaluate_features = Mock(
        spec=fs._feature_evaluator.evaluate_features, side_effect=["min", "mid", "max"]
    )
    data_splitter = Mock(DataSplitter)
    data_splitter.split_data = Mock(spec=data_splitter.split_data, return_value="data")

    res = fs._evaluate_min_mid_and_max_features(
        dataset, best_features, "split", data_splitter
    )

    assert res == ("min", "mid", "max")
    data_splitter.split_data.assert_called_with(dataset, "split", best_features["max"])
    assert fs._feature_evaluator.evaluate_features.call_count == 3


@pytest.mark.parametrize(
    "executor",
    [ProcessPoolExecutor(), Client().get_executor(), get_reusable_executor()],
)
def test_deferred_fit(executor):
    X = np.random.rand(10, 10)
    y = np.array([np.random.choice([0, 1]) for _ in range(10)])
    lr = LinearRegression()
    fs = FeatureSelector(
        n_outer=3,
        n_repetitions=2,
        random_state=0,
        estimator=lr,
        metric="MISS",
    )
    fitted_fs = fs.fit(X, y, executor=executor)
    assert fitted_fs is fs
    assert fs._selected_features
    assert fs.is_fit


def test_select_best_features(fs):
    fs._fetch_results = Mock(fs._fetch_results, return_value="results")
    fs._post_processor.select_features = Mock(
        fs._post_processor.select_features, return_value="features"
    )

    selected_features = fs._select_best_features("rep results")

    assert selected_features == "features"
    fs._fetch_results.assert_called_once_with("rep results")
    fs._post_processor.select_features.assert_called_once_with("results")


def test_get_feature_selection_results(fs, raw_results):
    fs._selected_features = [1, 2, 3]
    fs._raw_results = raw_results
    fs.is_fit = True
    fs._get_selected_feature_names = Mock(
        fs._get_selected_feature_names, return_value="sel_feat_names"
    )
    fs._get_validation_curves = Mock(
        fs._get_validation_curves, return_value={"total": [ScoreCurve(None, None)]}
    )

    fs_results = fs.get_feature_selection_results(["names"])

    assert isinstance(fs_results, FeatureSelectionResults)
    assert isinstance(fs_results.score_curves["total"][0], ScoreCurve)
    assert fs_results.selected_features == fs._selected_features
    assert fs_results.selected_features is not fs._selected_features
    assert fs_results.raw_results == fs._raw_results
    assert fs_results.raw_results is not fs._raw_results
    assert fs_results.selected_feature_names == "sel_feat_names"
    fs._get_selected_feature_names.assert_called_once_with(["names"])
    fs._get_validation_curves.assert_called_once()


def test_get_selected_features(fs):
    fs._selected_features = [1, 2, 3]

    selected_features = fs.get_selected_features()

    assert selected_features == fs._selected_features
    assert selected_features is not fs._selected_features


def test_get_selected_feature_names(fs, mosquito):
    X = mosquito.X[:, 0:10]
    y = np.array([1] + [0, 1] * 14)
    fs.fit(X, y)
    fs._selected_features = SelectedFeatures([0], [0], [0])
    selected_features = fs._selected_features
    feature_names = ["A", "B", "C", "D", "E", "F", "G", "H", "I", "L"]

    selected_features_names = fs._get_selected_feature_names(
        feature_names=feature_names
    )

    assert len(selected_features_names["min"]) == len(selected_features["min"])
    with pytest.raises(ValueError):
        fs._get_selected_feature_names(feature_names=["only-one-name"])


def test_export_average_feature_ranks(fs):
    df = pd.DataFrame()
    df.to_csv = Mock()
    fs.get_average_ranks_df = Mock(return_value=df)

    res = fs.export_average_feature_ranks("path", ["names"], True)

    assert res is df
    fs.get_average_ranks_df.assert_called_once_with(["names"], True)
    df.to_csv.assert_called_once_with("path")


def test_get_average_ranks_df(fs):
    fs.get_feature_selection_results = Mock(return_value="fs_results")
    fs._post_processor.make_average_ranks_df = Mock(return_value="ranks_df")

    ranks_df = fs.get_average_ranks_df(["names"], True)

    assert ranks_df == "ranks_df"
    fs.get_feature_selection_results.assert_called_once()
    fs._post_processor.make_average_ranks_df.assert_called_once_with(
        "fs_results", fs._n_features, ["names"], True
    )



from typing import List, Dict, Tuple
from scipy.stats import gmean
import numpy as np
import pandas as pd
from py_muvr.data_structures import (
    SelectedFeatures,
    ScoreCurve,
    FeatureEliminationResults,
    InnerLoopResults,
    FeatureSelectionRawResults,
    FeatureSelectionResults,
)
from py_muvr.utils import (
    average_ranks,
    average_scores,
    normalize_score,
    get_best_n_features,
)


class PostProcessor:
    """Contains several method that can process the results of the double CV loop
    to perform the actual feature selection and extract data for the plot utils.

    Parameters
    ----------
    robust_minimum : float
        maximum normalized-score value to be considered when computing the `min` and
        `max` selected features
    """

    def __init__(self, robust_minimum: float):
        self.robust_minimum = robust_minimum

    def select_features(self, results: FeatureSelectionRawResults) -> SelectedFeatures:
        """Select the best features for the three possible models (`min`, `mid`
        and `max`).
        The best features are chosen based on their average ranks in the three models
        across repetitions. The number of features are chosen based on the average score
        curve across repetitions and on `robust_minimum`.

        Parameters
        ----------
        results : FeatureSelectionRawResults
            results of the double CV feature selection

        Returns
        -------
        SelectedFeatures
            the best features for the three models
        """
        flat_results = [r for repetition in results for r in repetition]
        average_ranks_min = average_ranks([r.min_eval.ranks for r in flat_results])
        average_ranks_mid = average_ranks([r.mid_eval.ranks for r in flat_results])
        average_ranks_max = average_ranks([r.max_eval.ranks for r in flat_results])
        features_min, features_mid, features_max = self._compute_n_features(results)
        return SelectedFeatures(
            min=get_best_n_features(average_ranks_min, features_min),
            mid=get_best_n_features(average_ranks_mid, features_mid),
            max=get_best_n_features(average_ranks_max, features_max),
        )

    def _compute_n_features(self, results: FeatureSelectionRawResults):
        avg_rep_scores = self._get_repetition_avg_scores(results)
        avg_scores = average_scores(avg_rep_scores)
        norm_score = normalize_score(avg_scores)
        n_feats_close_to_min = [
            n for n, s in norm_score.items() if s <= self.robust_minimum
        ]
        max_feats = max(n_feats_close_to_min)
        min_feats = min(n_feats_close_to_min)
        mid_feats = gmean([max_feats, min_feats])
        mid_feats = min(norm_score.keys(), key=lambda x: abs(x - mid_feats))
        return min_feats, mid_feats, max_feats

    @staticmethod
    def _get_repetition_avg_scores(results: FeatureSelectionRawResults) -> List:
        avg_scores = []
        for outer_loops_results in results:
            scores = [ol.n_features_to_score_map for ol in outer_loops_results]
            avg_scores.append(average_scores(scores))
        return avg_scores

    def get_validation_curves(
        self, results: FeatureSelectionRawResults
    ) -> Dict[str, List[ScoreCurve]]:
        """Get validation curves (avg score vs number of features) for all outer
        loop iterations, and their average per repetition and overall. Every group
        of curves is stores in a list and packed in an output dictionary:

            {
                "outer_loops": list of outer loop curves,
                "repetitions": list of average ov outer loop curves across repetitions,
                "total": average of all curves,
            }

        Parameters
        ----------
        results : FeatureSelectionRawResults
            results of the double CV feature selection

        Returns
        -------
        Dict[str, List[ScoreCurve]]:
            List of score curves for all outer loop iterations, for all repetitions,
            and overall
        """
        flat_results = [r for repetition in results for r in repetition]
        outer_loop_scores = [r.n_features_to_score_map for r in flat_results]
        avg_scores_per_loop = self._get_repetition_avg_scores(results)
        avg_scores = average_scores(avg_scores_per_loop)
        return {
            "outer_loops": [self._score_to_score_curve(s) for s in outer_loop_scores],
            "repetitions": [self._score_to_score_curve(s) for s in avg_scores_per_loop],
            "total": [self._score_to_score_curve(avg_scores)],
        }

    @staticmethod
    def _score_to_score_curve(scores: Dict[int, float]) -> ScoreCurve:
        n_features, score_values = zip(*sorted(scores.items()))
        return ScoreCurve(n_features=n_features, scores=score_values)

    def process_feature_elim_results(self, raw_results: Dict[tuple, InnerLoopResults]):
        """Processes the feature elimination inner loop results to produce
        a score curve and a selection of best features from this curve. Results are
        packed in a FeatureEliminationResults object.

        Parameters
        ----------
        raw_results : Dict[tuple, InnerLoopResults]
            the results of the inner loops for every step of the recursive feature
            elimination

        Returns
        -------
        FeatureEliminationResults
            the processed results of the recursive feature elimination
        """
        n_feats_to_score = self._compute_score_curve(raw_results)
        best_features = self._select_best_outer_features(raw_results, n_feats_to_score)

        return FeatureEliminationResults(n_feats_to_score, best_features)

    @staticmethod
    def _compute_score_curve(
        elimination_results: Dict[tuple, InnerLoopResults]
    ) -> Dict[int, float]:
        avg_scores = {}
        for features, in_loop_res in elimination_results.items():
            n_feats = len(features)
            test_scores = [r.test_score for r in in_loop_res]
            avg_scores[n_feats] = np.average(test_scores)
        return avg_scores

    def _select_best_outer_features(
        self,
        elimination_results: Dict[tuple, InnerLoopResults],
        avg_scores: Dict[int, float],
    ) -> SelectedFeatures:
        n_to_features = self._compute_n_features_map(elimination_results)
        norm_score = normalize_score(avg_scores)
        n_feats_close_to_min = [
            n for n, s in norm_score.items() if s <= self.robust_minimum
        ]
        max_feats = max(n_feats_close_to_min)
        min_feats = min(n_feats_close_to_min)
        mid_feats = gmean([max_feats, min_feats])
        mid_feats = min(avg_scores.keys(), key=lambda x: abs(x - mid_feats))

        return SelectedFeatures(
            mid=n_to_features[mid_feats],
            min=n_to_features[min_feats],
            max=n_to_features[max_feats],
        )

    @staticmethod
    def _compute_n_features_map(
        elimination_results: Dict[tuple, InnerLoopResults]
    ) -> Dict[int, Tuple[int]]:
        n_to_features = {}
        for features in elimination_results.keys():
            n_feats = len(features)
            n_to_features[n_feats] = features
        return n_to_features

    def make_average_ranks_df(
        self,
        feature_selection_results: FeatureSelectionResults,
        n_features: int,
        feature_names: List[str] = None,
        exclude_unused_features: bool = True,
    ) -> pd.DataFrame:

        results_df = pd.DataFrame(
            index=np.arange(n_features), columns=["min", "mid", "max"]
        )

        for feature_set in ["min", "mid", "max"]:
            ranks = self._get_feature_ranks(
                feature_selection_results.raw_results, feature_set
            )
            res = pd.DataFrame(ranks).mean()
            results_df.loc[res.index, feature_set] = res.values

        if exclude_unused_features:
            results_df = results_df.dropna(how="all")

        if feature_names is not None:
            results_df.index = [feature_names[i] for i in results_df.index]

        return results_df

    @staticmethod
    def _get_feature_ranks(
        raw_results: FeatureSelectionRawResults, feature_set: str
    ) -> List[Dict[int, float]]:
        ranks = []
        for r in raw_results:
            for ol in r:
                ranks_raw_data = getattr(ol, feature_set + "_eval").ranks.get_data()
                ranks.append(ranks_raw_data)
        return ranks



import pytest
import numpy as np
from sklearn.model_selection import train_test_split
from py_muvr.models.pls import PLSClassifier, PLSRegressor
from sklearn.cross_decomposition import PLSRegression


@pytest.fixture
def pls_classifier():
    return PLSClassifier()


@pytest.fixture
def pls_regressor():
    return PLSRegressor()


def test_classifier_fit(pls_classifier, mosquito):
    fitted_classifier = pls_classifier.fit(mosquito.X, mosquito.y)
    assert fitted_classifier is pls_classifier
    assert fitted_classifier.coef_ is not None
    assert fitted_classifier.feature_importances_ is not None
    assert len(fitted_classifier.feature_importances_) == mosquito.X.shape[1]


def test_classifier_prediction(pls_classifier, mosquito):
    fitted_classifier = pls_classifier.fit(mosquito.X, mosquito.y)
    y_pred = pls_classifier.predict(mosquito.X)
    assert y_pred is not None
    assert set(y_pred) == set(mosquito.y)
    assert y_pred.shape == mosquito.y.shape
    y_val = set(mosquito.y).pop()
    binary_y = (mosquito.y == y_val).astype(float)
    fitted_classifier = pls_classifier.fit(mosquito.X, binary_y)
    y_pred = pls_classifier.predict(mosquito.X)
    assert set(y_pred) == {0, 1}


def test_regressor_fit(pls_regressor):
    X = np.random.rand(10, 10)
    y = np.random.rand(10)
    sklearn_regressor = PLSRegression().fit(X, y)
    assert pls_regressor.fit(X, y)
    assert pls_regressor.fit(X[:, 0:1], y)
    with pytest.raises(ValueError):
        sklearn_regressor.fit(X[:, 0:1], y)


def test_regressor_predict(pls_regressor):
    X = np.random.rand(10, 10)
    y = np.random.rand(10)
    sklearn_regressor = PLSRegression().fit(X, y)
    pls_regressor.fit(X, y)
    y_pred = pls_regressor.predict(X)
    assert y_pred.shape == y.shape
    assert np.all(y_pred == sklearn_regressor.predict(X).ravel())



import pytest

import numpy as np

from py_muvr.exceptions import NotFitException
from py_muvr.consensus_model import ConsensusModel
from py_muvr.feature_selector import FeatureSelector
from py_muvr.models import ScikitLearnEstimator


@pytest.fixture
def fitted_selector(dataset):
    return FeatureSelector(
        n_outer=3,
        n_repetitions=2,
        metric="MISS",
        estimator="PLSC",
        random_state=0,
    ).fit(dataset.X, dataset.y)


def test_consensus_model_init(fitted_selector):
    fs = FeatureSelector(
        n_outer=3,
        n_repetitions=2,
        metric="MISS",
        estimator="PLSC",
        random_state=0,
    )
    with pytest.raises(NotFitException):
        _ = ConsensusModel(fs, "min")
    with pytest.raises(ValueError):
        _ = ConsensusModel(fitted_selector, "yo")
    assert ConsensusModel(fitted_selector, "min")
    assert ConsensusModel(fitted_selector, "max")
    assert ConsensusModel(fitted_selector, "mid")


def test_predict(fitted_selector, dataset):
    cm = ConsensusModel(fitted_selector, "min")
    y_pred = cm.predict(dataset.X)
    assert y_pred.shape == dataset.y.shape
    assert set(y_pred.tolist()) == set(dataset.y.tolist())


def test_extract_evaluators(fitted_selector):
    cm = ConsensusModel(fitted_selector, "min")
    for model in cm._models:
        assert isinstance(model, ScikitLearnEstimator)


def test_extract_feature_sets(fitted_selector):
    cm = ConsensusModel(fitted_selector, "min")
    assert (
        len(cm._feature_sets) == fitted_selector.n_repetitions * fitted_selector.n_outer
    )
    for features in cm._feature_sets:
        assert len(features) <= fitted_selector._n_features
        assert max(features) <= fitted_selector._n_features - 1
        assert min(features) >= 0


def test_predict_with_regressor(freelive):
    fs = FeatureSelector(
        n_outer=3,
        n_repetitions=2,
        metric="neg_mean_squared_error",
        estimator="PLSR",
        random_state=0,
    ).fit(freelive.X[:, :100], freelive.y)
    cm = ConsensusModel(fs, "mid")

    y_pred = cm.predict(freelive.X[:, :200])

    assert y_pred.shape == freelive.y.shape
    assert all(y_pred >= 0)
    assert np.floor(y_pred.mean()) == np.floor(freelive.y.mean())



import pytest
import numpy as np
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import Normalizer
from sklearn.svm import SVC
from py_muvr.data_structures import FeatureRanks
from py_muvr.feature_evaluator import FeatureEvaluator
from py_muvr.models.sklearn_estimator import ScikitLearnEstimator
from py_muvr.data_splitter import DataSplitter


@pytest.fixture
def feature_evaluator():
    fe = FeatureEvaluator(
        estimator="RFC",
        metric="MISS",
        random_state=0,
    )
    fe.set_n_initial_features(12)
    return fe


def test_feature_evaluator(feature_evaluator):
    assert feature_evaluator
    assert feature_evaluator._estimator
    assert hasattr(feature_evaluator._metric, "__call__")
    assert feature_evaluator._n_initial_features == 12


def test_evaluate_features(dataset):
    pipeline = Pipeline(
        [("normalizer", Normalizer()), ("model", SVC(kernel="linear", random_state=0))]
    )
    fe = FeatureEvaluator(
        estimator=pipeline,
        metric="MISS",
        random_state=0,
    )
    fe.set_n_initial_features(12)
    ds = DataSplitter(n_outer=5, n_inner=4, random_state=0, input_data=dataset)
    split = ds.iter_outer_splits().__next__()
    evaluation_data = ds.split_data(dataset, split)
    evaluation = fe.evaluate_features(evaluation_data, [0, 4, 6])
    assert evaluation
    assert evaluation.test_score >= 0
    assert evaluation.ranks
    assert isinstance(evaluation.ranks, FeatureRanks)
    assert evaluation.ranks.n_feats == 12
    assert evaluation.ranks[0]
    assert evaluation.ranks[1]
    with pytest.raises(ValueError):
        _ = evaluation.ranks[100]


class EstimatorWithFI:
    feature_importances_ = None

    def set_params(*args, **kwargs):
        pass


class EstimatorWithCoeffs:
    feature_coef_ = None

    def set_params(*args, **kwargs):
        pass


@pytest.mark.parametrize(
    ("estimator", "attribute", "values"),
    [
        (
            EstimatorWithFI,
            "feature_importances_",
            np.array([0.4, 0.05, 0.4, 0, 0, 0.15]),
        ),
        (EstimatorWithCoeffs, "coef_", np.array([[-0.4, -0.05, 0.4, 0, 0, 0.15]])),
    ],
)
def test_get_feature_rank(estimator, attribute, values, dataset, feature_evaluator):
    setattr(estimator, attribute, values)
    model = ScikitLearnEstimator(estimator, None)
    features = [1, 7, 8, 9, 10, 11]
    ranks = feature_evaluator._get_feature_ranks(model, features)

    assert ranks[1] == 1.5
    assert ranks[7] == 4
    assert ranks[8] == 1.5
    assert ranks[9] == 5.5
    assert ranks[10] == 5.5
    assert ranks[11] == 3



import logging
from typing import Iterable, Tuple, List
from concurrent.futures import Executor
import numpy as np
from scipy.stats import rankdata, normaltest
import progressbar
from py_muvr.utils import compute_t_student_p_value, mute_loggers
from py_muvr.data_structures.data_types import NumpyArray
from py_muvr.data_structures.data_models import FeatureSelectionResults
from py_muvr.feature_selector import FeatureSelector

logger = logging.getLogger(__name__)


class PermutationTest:
    """ "Implements a permutation test of the py_muvr feature selection.
    The tests fits the input `feature_selector` `n_permutations` times.
    The target `y` is scrambled at every iteration so that the original
    feature selection results can be contrasted against the results of the
    randomized ones. Statistical significance is addressed using a t-Test.

    The results of the randomized fit can be inspected using
    `compute_permutation_scores`.
    If the scores returned by the permutations are very non-normally distributed,
    it's probably better to invoke `compute_p_values` with `ranks=True`, to mitigate
    the violation of the normality assumption needed to perform a meaningful t-Test.

    NOTE: for efficiency, if the input feature selector has been fit already,
    it will not be fit again against the original X, y pair.

    Parameters
    ----------
    feature_selector: FeatureSelector
        the FeatureSelector instance to be used
    n_permutations: int
        number of times the fit on shuffled `y`is repeated

    Examples
    --------
    >>> fs = FeatureSelector(n_outer=6, metric="MISS", estimator="PLSC")
    >>> pt = PermutationTest(fs, 10)
    >>> pt.fit(X, y)
    >>> pt.compute_p_values(model="min")

    """

    def __init__(self, feature_selector: FeatureSelector, n_permutations: int = 20):
        self.n_permutations = n_permutations
        self._fs = feature_selector
        self.res = None
        self.res_perm = None

    @mute_loggers(loggers=["py_muvr.feature_selector", "py_muvr.models.pls"])
    def fit(
        self,
        X: NumpyArray,
        y: NumpyArray,
        groups: NumpyArray = None,
        executor: Executor = None,
    ):
        logger.info("Running permutation test for %s permutations", self.n_permutations)
        y_idx = np.arange(y.size)

        if not self._fs.is_fit:
            self._fs.fit(X, y, groups=groups)
        self.res = self._fs.get_feature_selection_results()

        fs_perm = self._copy_feature_selector(self._fs)
        self.res_perm = []
        for _ in progressbar.progressbar(range(self.n_permutations)):
            np.random.shuffle(y_idx)
            y_perm = y[y_idx]
            fs_perm.fit(X, y_perm, groups=groups, executor=executor)
            self.res_perm.append(fs_perm.get_feature_selection_results())
        logger.info("Finished permutation test. Storing results in self.res_perm")
        return self

    def compute_permutation_scores(self, model: str) -> Tuple[float, List[float]]:
        """Compute the permutation tests scores for input model. `model` must be one of
        `"min"`, `"mid"` or `"max"`. It returns a tuple. The first element is the score
        of the original feature selection, the second element is a list of scores, one
        for each permutation

        Parameters
        ----------
        model : str
            one of `"min"`, `"mid"` or `"max"` models

        Returns
        -------
        tuple
            the original feature selection score and the scores for every permutation

        Raises
        ------
        RuntimeError
            if `fit` is not called first
        ValueError
            if `model` is not one of the three values allowed
        """
        if not self.res:
            raise RuntimeError("Call fit method first")
        if model not in {"min", "mid", "max"}:
            raise ValueError("Input model must be one of 'min', 'mid' or 'max'")
        score = self._compute_model_score(model, self.res)
        score_perm = [self._compute_model_score(model, r) for r in self.res_perm]
        return score, score_perm

    def compute_p_values(self, model: str, ranks: bool = False) -> float:
        """Compute the p-value relative to the original feature selection score for
        the input `model` based on the results of the various permutations.
        If the permutation scores are not normal, `ranks` is automatically set to True.

        Parameters
        ----------
        model : str
            one of `"min"`, `"mid"` or `"max"` models
        ranks : bool, optional
            whether to perform the significance tests on the ranks rather than on
            the actual scores, by default False

        Returns
        -------
        float
            the permutation test p-value
        """
        if model not in {"min", "mid", "max"}:
            raise ValueError("Input model must be one of 'min', 'mid' or 'max'")
        score, score_perm = self.compute_permutation_scores(model)

        if not ranks and len(score_perm) > 7:  # or `normaltest` would raise an error
            test = normaltest(score_perm)
            if test.pvalue < 0.05:
                logger.warning(
                    "%s %s",
                    "The permutation scores don't seem to be normally distributed.",
                    "Setting ranks to True (non-parametric test)",
                )
                ranks = True

        if ranks:
            score, score_perm = self._rank_data(score, score_perm)

        p_values = compute_t_student_p_value(score, score_perm)
        return p_values

    @staticmethod
    def _compute_model_score(
        model: str, feats_and_scores: FeatureSelectionResults
    ) -> float:
        selected_features = feats_and_scores.selected_features
        total_score_curve = feats_and_scores.score_curves["total"][0]
        features = getattr(selected_features, model)
        n_feats = len(features)
        if model == "mid":
            # mid is the geometric mean, the score curve might not contain it
            n_feats = min(total_score_curve.n_features, key=lambda x: abs(x - n_feats))
        idx = total_score_curve.n_features.index(n_feats)
        return total_score_curve.scores[idx]

    @staticmethod
    def _rank_data(sample: float, population: Iterable):
        ranks = rankdata([sample] + list(population))
        return ranks[0], ranks[1:]

    @staticmethod
    def _copy_feature_selector(fs: FeatureSelector) -> FeatureSelector:
        return FeatureSelector(
            n_outer=fs.n_outer,
            metric=fs.metric,
            estimator=fs.estimator,
            features_dropout_rate=fs.features_dropout_rate,
            robust_minimum=fs.robust_minimum,
            n_inner=fs.n_inner,
            n_repetitions=fs.n_repetitions,
            random_state=(
                None if fs.random_state is None else fs.random_state.get_state()[1][0]
            ),
        )

    def __repr__(self):
        return f"PermutationTest(n_permutations={self.n_permutations})"



from typing import List
from sklearn.metrics import (
    r2_score,
    median_absolute_error,
    max_error,
    mean_absolute_error,
    mean_squared_error,
    mean_squared_log_error,
    mean_poisson_deviance,
    mean_gamma_deviance,
    accuracy_score,
    f1_score,
    precision_score,
    recall_score,
    balanced_accuracy_score,
    explained_variance_score,
    jaccard_score,
)


class SklearnMetricsWrapper:
    def __init__(self, score_function, greater_is_better=True, **kwargs):
        self.score_func = score_function
        self.kwargs = kwargs
        self.sign = 1 if greater_is_better else -1
        self.greater_is_better = greater_is_better

    def __call__(self, y_true, y_pred):
        return self.sign * self.score_func(y_true, y_pred, **self.kwargs)

    def __repr__(self):
        return (
            f"SklearnMetricsWrapper(score_function={self.score_func}, "
            + f"greater_is_better={self.greater_is_better})"
        )


_SKLEARN_REGRESSION_METRICS = {
    "explained_variance": SklearnMetricsWrapper(explained_variance_score),
    "r2": SklearnMetricsWrapper(r2_score),
    "max_error": SklearnMetricsWrapper(max_error, greater_is_better=False),
    "neg_median_absolute_error": SklearnMetricsWrapper(
        median_absolute_error, greater_is_better=False
    ),
    "neg_mean_absolute_error": SklearnMetricsWrapper(
        mean_absolute_error, greater_is_better=False
    ),
    "neg_mean_squared_error": SklearnMetricsWrapper(
        mean_squared_error, greater_is_better=False
    ),
    "neg_mean_squared_log_error": SklearnMetricsWrapper(
        mean_squared_log_error, greater_is_better=False
    ),
    "neg_root_mean_squared_error": SklearnMetricsWrapper(
        mean_squared_error, greater_is_better=False, squared=False
    ),
    "neg_mean_poisson_deviance": SklearnMetricsWrapper(
        mean_poisson_deviance, greater_is_better=False
    ),
    "neg_mean_gamma_deviance": SklearnMetricsWrapper(
        mean_gamma_deviance, greater_is_better=False
    ),
}

_SKLEARN_CLASSIFICATION_METRICS = {
    "accuracy": SklearnMetricsWrapper(accuracy_score),
    "balanced_accuracy": SklearnMetricsWrapper(balanced_accuracy_score),
    "precision": SklearnMetricsWrapper(precision_score, average="binary"),
    "precision_macro": SklearnMetricsWrapper(
        precision_score, pos_label=None, average="macro"
    ),
    "precision_micro": SklearnMetricsWrapper(
        precision_score, pos_label=None, average="micro"
    ),
    "recall": SklearnMetricsWrapper(recall_score, average="binary"),
    "recall_macro": SklearnMetricsWrapper(
        recall_score, pos_label=None, average="macro"
    ),
    "recall_micro": SklearnMetricsWrapper(
        recall_score, pos_label=None, average="micro"
    ),
    "f1": SklearnMetricsWrapper(f1_score, average="binary"),
    "f1_macro": SklearnMetricsWrapper(f1_score, pos_label=None, average="macro"),
    "f1_micro": SklearnMetricsWrapper(f1_score, pos_label=None, average="micro"),
    "jaccard": SklearnMetricsWrapper(jaccard_score, average="binary"),
    "jaccard_macro": SklearnMetricsWrapper(
        jaccard_score, pos_label=None, average="macro"
    ),
    "jaccard_micro": SklearnMetricsWrapper(
        jaccard_score, pos_label=None, average="micro"
    ),
}


SKLEARN_METRICS = _SKLEARN_CLASSIFICATION_METRICS.copy()
SKLEARN_METRICS.update(_SKLEARN_REGRESSION_METRICS)


def get_supported_classification_metrics() -> List[str]:
    return list(_SKLEARN_CLASSIFICATION_METRICS.keys())


def get_supported_regression_metrics() -> List[str]:
    return list(_SKLEARN_REGRESSION_METRICS.keys())



from __future__ import annotations

import logging
from concurrent.futures import Executor, Future
from copy import deepcopy
from typing import Union, List, Dict, Tuple

import numpy as np
import pandas as pd
import progressbar
from numpy.random import RandomState

from py_muvr.data_structures import (
    InputDataset,
    SelectedFeatures,
    OuterLoopResults,
    InnerLoopResults,
    Split,
    FeatureEvaluationResults,
    MetricFunction,
    InputEstimator,
    NumpyArray,
    FeatureSelectionRawResults,
    FeatureSelectionResults,
)
from py_muvr.feature_evaluator import FeatureEvaluator
from py_muvr.data_splitter import DataSplitter
from py_muvr.models import Estimator
from py_muvr.post_processor import PostProcessor
from py_muvr.utils import get_best_n_features, average_ranks
from py_muvr.exceptions import NotFitException
from py_muvr.sync_executor import SyncExecutor

Repetition = List[Union[OuterLoopResults, Future]]
log = logging.getLogger(__name__)


class FeatureSelector:
    """Feature selection based on double cross validation and iterative feature
    elimination.
    This class is based on the feature selection algorithm proposed in
    "Variable selection and validation in multivariate modelling", Shi L. et al.,
    Bioinformatics 2019
    https://academic.oup.com/bioinformatics/article/35/6/972/5085367

    Perform recursive feature selection using nested cross validation to select
    the optimal number of features explaining the relationship between `X` and `y`.
    The algorithm outputs three sets of features, that can be accessed via
    self.get_selected_features.

    1. `min`: is the minimum number of feature that gives good predictive power
    2. `max`: is the maximum number of feature that gives good predictive power
    3. `mid`: is the set of features to build a model using a number of feature
        that is the geometric mean of the minimum and the maximum number of features

    The structure of the nested CV loops is the following:
    - Repetitions
        - Outer CV Loops
            - Iterative Feature removal
                - Inner CV loops

    The inner loop are used to understand which feature to drop at each
    iteration removal.
    For each outer loop element, we have a score curve linking the fitness to the
    number of features, and average ranks for each variable.
    From the average of these curves, the number of variables for each "best" model
    (MIN, MID and MAX) are extracted and the feature rank of the best models
    are computed.

    Averaging the results and the feature importances across repetitions, we select
    the final set of features.

    The actual feature selection is performed by the `fit` method which implements the
    algorithm described in the original paper and developed in
    https://gitlab.com/CarlBrunius/MUVR/-/tree/master/R

    For additional informations about the algorithm, please check the original
    paper linked above.

    Parameters
    ----------
    n_outer: int
        number of outer CV folds
    metric: Union[str, MetricFunction]
        metric to be used to assess estimator goodness
    estimator: Union[str, InputEstimator]
        estimator to be used for feature elimination
    features_dropout_rate: float
        fraction of features to drop at each elimination step
    robust_minimum: float
        maximum normalized-score value to be considered when computing the `min` and
        `max` selected features
    n_inner: int
        number of inner CV folds, by default n_outer - 1
    n_repetitions: int
        number of repetitions of the double CV loops, by default 8
    random_state: int
        pass an int for reproducible output, by default None
    """

    def __init__(
        self,
        n_outer: int,
        metric: Union[str, MetricFunction],
        estimator: Union[str, InputEstimator],
        features_dropout_rate: float = 0.05,
        robust_minimum: float = 0.05,
        n_inner: int = None,
        n_repetitions: int = 8,
        random_state: int = None,
    ):
        self.n_outer = n_outer
        self.metric = metric
        self.estimator = estimator
        self.features_dropout_rate = features_dropout_rate
        self.robust_minimum = robust_minimum
        self.n_inner = self._set_n_inner(n_inner)
        self.n_repetitions = n_repetitions
        self.random_state = None if random_state is None else RandomState(random_state)

        self.is_fit = False
        self._keep_fraction = 1 - features_dropout_rate
        self._n_features = None
        self._selected_features = None
        self._raw_results = None
        self._minimum_features = 1

        self._feature_evaluator = FeatureEvaluator(estimator, metric, random_state)
        self._post_processor = PostProcessor(robust_minimum)

    @property
    def raw_results(self):
        if self.is_fit:
            return self._raw_results
        else:
            raise NotFitException("The feature selector is not fit yet")

    def _set_n_inner(self, n_inner: Union[int, None]) -> int:
        if not n_inner:
            log.info("Parameter n_inner is not specified, setting it to n_outer - 1")
            n_inner = self.n_outer - 1
        return n_inner

    def fit(
        self,
        X: NumpyArray,
        y: NumpyArray,
        groups: NumpyArray = None,
        executor: Executor = None,
    ) -> FeatureSelector:
        """
        Implements the double CV feature selection algorithm. The method returns
        the same FeatureSelector. If the samples are correlated, the `group` vector
        can be used to encode arbitrary domain specific stratifications of the
        samples as integers (e.g. patient_id, year of collection, etc.). If group
        is not provided the samples are assumed to be i. i. d. variables.
        To parallelize the CV repetition, an `executor` can be provided to split
        the computation across processes or cluster nodes. So far, `loky` (joblib),
        `dask`, and `concurrent` Executors are tested.

        Parameters
        ----------
        X : NumpyArray
            Predictor variables as numpy array
        y : NumpyArray
            Response vector (Dependent variable).
        groups : NumpyArray, optional
            Group labels for the samples used while splitting the dataset
            into train/test set, by default None
        executor : Executor, optional
            executor instance for parallel computing, by default None

        Returns
        -------
        FeatureSelector
            the fit feature selector
        """

        if executor is None:
            executor = SyncExecutor()

        size, n_features = X.shape
        groups = self._get_groups(groups, size)
        input_data = InputDataset(X=X, y=y, groups=groups)
        self._feature_evaluator.set_n_initial_features(n_features)

        log.info(
            f"Running {self.n_repetitions} repetitions and"
            f" {self.n_outer} outer loops using "
            f"executor {executor.__class__.__name__}."
        )

        repetition_results = []

        log.info("Scheduling tasks...")
        Progressbar = self._make_progress_bar()
        with Progressbar(max_value=self.n_repetitions * self.n_outer) as b:
            progress = 0
            b.update(progress)
            for _ in range(self.n_repetitions):
                data_splitter = DataSplitter(
                    self.n_outer,
                    self.n_inner,
                    input_data,
                    self.random_state,
                )

                outer_loop_results = []
                for outer_split in data_splitter.iter_outer_splits():
                    outer_loop_result = self._deferred_run_outer_loop(
                        input_data,
                        outer_split,
                        executor=executor,
                        data_splitter=data_splitter,
                    )
                    outer_loop_results.append(outer_loop_result)
                    progress += 1
                    b.update(progress)

                repetition_results.append(outer_loop_results)

        self._selected_features = self._select_best_features(repetition_results)
        log.info("Finished feature selection.")
        self._n_features = input_data.n_features
        self.is_fit = True
        return self

    @staticmethod
    def _get_groups(groups: NumpyArray, size: int) -> NumpyArray:
        if groups is None:
            log.info("Groups parameter is not specified: independent samples assumed")
            groups = np.arange(size)
        return groups

    def _deferred_run_outer_loop(
        self,
        input_data: InputDataset,
        outer_split: Split,
        data_splitter: DataSplitter,
        executor: Executor,
    ) -> Union[Future, OuterLoopResults]:
        if executor is None:
            return self._run_outer_loop(input_data, outer_split, data_splitter)
        return executor.submit(
            self._run_outer_loop, input_data, outer_split, data_splitter
        )

    def _run_outer_loop(
        self,
        input_data: InputDataset,
        outer_split: Split,
        data_splitter: DataSplitter,
    ) -> OuterLoopResults:

        feature_elimination_results = {}
        feature_set = list(range(input_data.n_features))

        while len(feature_set) >= self._minimum_features:
            inner_results = []

            for inner_split in data_splitter.iter_inner_splits(outer_split):
                inner_loop_data = data_splitter.split_data(
                    input_data, inner_split, feature_set
                )

                feature_evaluation_results = self._feature_evaluator.evaluate_features(
                    inner_loop_data, feature_set
                )

                inner_results.append(feature_evaluation_results)

            feature_elimination_results[tuple(feature_set)] = inner_results
            feature_set = self._remove_features(feature_set, inner_results)

        outer_loop_results = self._create_outer_loop_results(
            feature_elimination_results, input_data, outer_split, data_splitter
        )

        return outer_loop_results

    def _remove_features(
        self, features: List[int], results: InnerLoopResults
    ) -> List[int]:
        features_to_keep = int(np.floor(len(features) * self._keep_fraction))
        features = self._select_n_best(results, features_to_keep)
        return features

    @staticmethod
    def _select_n_best(inner_loop_result: InnerLoopResults, keep_n: int) -> List[int]:
        if keep_n < 1:
            return []
        ranks = [r.ranks for r in inner_loop_result]
        avg_ranks = average_ranks(ranks)
        return get_best_n_features(avg_ranks, keep_n)

    def _create_outer_loop_results(
        self,
        raw_feature_elim_results: Dict[tuple, InnerLoopResults],
        input_data: InputDataset,
        outer_split: Split,
        data_splitter: DataSplitter,
    ) -> OuterLoopResults:
        feature_elimination_results = self._post_processor.process_feature_elim_results(
            raw_feature_elim_results
        )
        min_eval, mid_eval, max_eval = self._evaluate_min_mid_and_max_features(
            input_data,
            feature_elimination_results.best_features,
            outer_split,
            data_splitter,
        )
        outer_loop_results = OuterLoopResults(
            min_eval=min_eval,
            mid_eval=mid_eval,
            max_eval=max_eval,
            n_features_to_score_map=feature_elimination_results.n_features_to_score_map,
        )
        return outer_loop_results

    def _evaluate_min_mid_and_max_features(
        self,
        input_data: InputDataset,
        best_features: SelectedFeatures,
        split: Split,
        data_splitter: DataSplitter,
    ) -> Tuple[
        FeatureEvaluationResults, FeatureEvaluationResults, FeatureEvaluationResults
    ]:
        min_feats = best_features["min"]
        mid_feats = best_features["mid"]
        max_feats = best_features["max"]

        data_min_feats = data_splitter.split_data(input_data, split, min_feats)
        data_mid_feats = data_splitter.split_data(input_data, split, mid_feats)
        data_max_feats = data_splitter.split_data(input_data, split, max_feats)

        min_eval = self._feature_evaluator.evaluate_features(data_min_feats, min_feats)
        mid_eval = self._feature_evaluator.evaluate_features(data_mid_feats, mid_feats)
        max_eval = self._feature_evaluator.evaluate_features(data_max_feats, max_feats)

        return min_eval, mid_eval, max_eval

    def _select_best_features(
        self, repetition_results: FeatureSelectionRawResults
    ) -> SelectedFeatures:
        self._raw_results = self._fetch_results(repetition_results)
        selected_features = self._post_processor.select_features(self._raw_results)
        return selected_features

    def _fetch_results(
        self, results: FeatureSelectionRawResults
    ) -> FeatureSelectionRawResults:

        log.info("Retrieving results...")
        Progressbar = self._make_progress_bar()
        with Progressbar(max_value=self.n_repetitions * self.n_outer) as b:
            progress = 0
            b.update(progress)

            fetched_results = []
            for repetition in results:
                ol_results = []

                for outer_loop_result in repetition:
                    fetched_outer_loop = outer_loop_result.result()
                    ol_results.append(fetched_outer_loop)

                    progress += 1
                    b.update(progress)

                fetched_results.append(ol_results)
        return fetched_results

    def get_feature_selection_results(
        self, feature_names: List[str] = None
    ) -> FeatureSelectionResults:
        """
        Retrieve the feature selection results in a single data structure. This object
        contains the attributes:
        - raw_results: selection results before processing
        - selected_features: 0-based integer indices for min, mid and max feature sets
        - score_curves: validation curves of n_feats vs score
        - selected_feature_names: list of feature names if the parameter is used.

        Parameters
        ----------
        feature_names : List[str], optional
            the name of every feature, by default None

        Returns
        -------
        FeatureSelectionResults
            The results obtained from running the algorithm

        Raises
        ------
        NotFitException
            if the `fit` method was not called successfully already
        """
        if not self.is_fit:
            raise NotFitException("The feature selector is not fit yet")

        return FeatureSelectionResults(
            raw_results=deepcopy(self._raw_results),
            selected_features=self.get_selected_features(),
            score_curves=self._get_validation_curves(),
            selected_feature_names=self.get_selected_features(feature_names),
        )

    def _get_selected_feature_names(
        self, feature_names: Union[None, List[str]]
    ) -> Union[None, SelectedFeatures]:
        if feature_names is None:
            return feature_names

        if len(feature_names) != self._n_features:
            raise ValueError(
                f"feature_names provided should contain {self._n_features} elements"
            )
        min_names = [feature_names[f] for f in self._selected_features["min"]]
        mid_names = [feature_names[f] for f in self._selected_features["mid"]]
        max_names = [feature_names[f] for f in self._selected_features["max"]]

        selected_feature_names = SelectedFeatures(
            min=min_names,
            max=max_names,
            mid=mid_names,
        )
        return selected_feature_names

    def get_selected_features(self, feature_names: List[str] = None):
        """Retrieve the selected feature for the three models. Features are normally
        returned as 0-based integer indices representing the columns of the input
        predictor variables (X), however if a list of feature names is provided via
        `feature_names`, the feature names are returned instead.

        Parameters
        ----------
        feature_names : List[str], optional
            the name of every feature, by default None

        Returns
        -------
        SelectedFeatures
            The features selected by the double CV loops

        Raises
        ------
        NotFitException
            if the `fit` method was not called successfully already
        """
        if feature_names is None:
            return deepcopy(self._selected_features)
        else:
            return self._get_selected_feature_names(feature_names)

    def _get_validation_curves(self) -> Dict[str, List]:
        return self._post_processor.get_validation_curves(self._raw_results)

    def __repr__(self):
        fs = (
            f"FeatureSelector("
            f"repetitions={self.n_repetitions},"
            f" n_outer={self.n_outer},"
            f" n_inner={self.n_inner},"
            f" feature_dropout_rate={self.features_dropout_rate},"
            f" is_fit={self.is_fit})"
        )

        return fs

    @staticmethod
    def _make_progress_bar():
        if logging.getLogger(__name__).getEffectiveLevel() > logging.INFO:
            return progressbar.NullBar
        return progressbar.ProgressBar

    def export_average_feature_ranks(
        self,
        output_path: str,
        feature_names: List[str] = None,
        exclude_unused_features: bool = True,
    ) -> pd.DataFrame:
        """
        Creates and saves dataframe from the feature selection results. This dataframe contains
        the columns 'mid', 'mid', and 'max', the indices are the features and the values
        are the average rank across repetitions and outer loops.

        Parameters
        ----------
        output_path: str
            Path where to save the csv dataframe
        feature_names: List[str]
            The name of every feature, by default None
        exclude_unused_features: bool
            Whether to remove the features that werent selected or not.

        Returns
        -------
        pd.DataFrame:
            Pandas dataframe containing the average feature ranks

        """
        ranks_df = self.get_average_ranks_df(feature_names, exclude_unused_features)

        ranks_df.to_csv(output_path)
        return ranks_df

    def get_average_ranks_df(
        self,
        feature_names: List[str] = None,
        exclude_unused_features: bool = True,
    ):
        """
        Creates a dataframe from the feature selection results. This dataframe contains
        the columns 'mid', 'mid', and 'max', the indices are the features and the values
        are the average rank across repetitions and outer loops.

        Parameters
        ----------
        feature_names: List[str]
            The name of every feature, by default None
        exclude_unused_features: bool
            Whether to remove the features that werent selected or not.

        Returns
        -------
        pd.DataFrame:
            Pandas dataframe containing the average feature ranks

        """
        results = self.get_feature_selection_results()
        return self._post_processor.make_average_ranks_df(
            results, self._n_features, feature_names, exclude_unused_features
        )



from unittest.mock import Mock
import pytest
from distributed import Client

from py_muvr import FeatureSelector
from py_muvr.plot_utils import (
    plot_validation_curves,
    plot_feature_rank,
    plot_permutation_scores,
)
from py_muvr.permutation_test import PermutationTest


@pytest.fixture
def permutation_test():
    pt = Mock(PermutationTest)
    pt.compute_permutation_scores = Mock(
        spec=pt.compute_permutation_scores, return_value=(1, list(range(2, 1000)))
    )
    pt.compute_p_values = Mock(spec=pt.compute_p_values, return_value=0.01)
    return pt


def test_plot_validation_curves(fs_results):
    ax = plot_validation_curves(fs_results)
    assert ax


@pytest.mark.parametrize("model", ["min", "max", "mid"])
def test_plot_feature_rank(fs_results, model):
    fig = plot_feature_rank(fs_results, model)
    assert fig


def test_plot_feature_rank_error(fs_results):
    with pytest.raises(ValueError):
        fig = plot_feature_rank(fs_results, "yo")


def test_plot_permutiation_scores(permutation_test):
    fig = plot_permutation_scores(permutation_test, "min")
    assert fig



from py_muvr.data_structures import SelectedFeatures


def test_n_features(dataset):
    assert dataset.n_features == 12


def test_input_data_slice(dataset):
    assert dataset[:5, 3:7].X.shape == (5, 4)
    assert dataset[[1, 2, 5], [3, 4, 7]].X.shape == (3, 3)
    assert dataset[1:3, :].X.shape == (2, 12)
    assert dataset[:, :].X.shape == dataset.X.shape


def test_selected_features():
    sf = SelectedFeatures(min="min", mid="mid", max="max")

    assert sf["mid"]



from typing import Any

from sklearn.base import BaseEstimator
from sklearn.ensemble import RandomForestClassifier
from sklearn.pipeline import Pipeline
from xgboost import XGBClassifier

from py_muvr.data_structures import RandomState
from py_muvr.models.estimator import Estimator
from py_muvr.models.pls import PLSClassifier, PLSRegressor
from py_muvr.models.sklearn_estimator import ScikitLearnPipeline, ScikitLearnEstimator


class ESTIMATORS:
    RFC = "RFC"
    XGBC = "XGBC"
    PLSC = "PLSC"
    PLSR = "PLSR"


def make_estimator(estimator: Any, random_state: RandomState) -> Estimator:
    """Factory of Estimator classes based on an input `estimator`. So far, this method
    supports input strings or scikitlearn-like objects (e.g. xgboost.XGBClassifier).
    If a string is provided it must be one of `estimator.ESTIMATORS`

    Parameters
    ----------
    estimator : Any
        Input estimator. It can be a scikitlearn model or a string
    random_state : RandomState
        random state instance to control reproducibility

    Returns
    -------
    Estimator
        An instance of the Estimator class

    Raises
    ------
    ValueError
        if estimator does not comply with the aforementioned types
    """
    if isinstance(estimator, str):
        return _make_estimator_from_string(estimator, random_state)
    if isinstance(estimator, Pipeline):
        return ScikitLearnPipeline(estimator, random_state)
    if isinstance(estimator, BaseEstimator):
        return ScikitLearnEstimator(estimator, random_state)
    raise ValueError("Unknown type of estimator")


def _make_estimator_from_string(estimator: str, random_state: RandomState) -> Estimator:
    if estimator == ESTIMATORS.RFC:
        rfc = RandomForestClassifier(n_estimators=150)
        return ScikitLearnEstimator(rfc, random_state)
    if estimator == ESTIMATORS.PLSC:
        plsc = PLSClassifier()
        return ScikitLearnEstimator(plsc, random_state)
    if estimator == ESTIMATORS.XGBC:
        xgbc = XGBClassifier()
        return ScikitLearnEstimator(xgbc, random_state)
    if estimator == ESTIMATORS.PLSR:
        plsr = PLSRegressor()
        return ScikitLearnEstimator(plsr, random_state)
    raise ValueError("Unknown type of estimator")



from typing import List, Dict, Iterable, Callable
import logging
import pandas as pd
import numpy as np
import scipy.stats
from py_muvr.data_structures import FeatureRanks


def average_scores(scores: List[Dict]) -> Dict[int, float]:
    avg_score = pd.DataFrame(scores).fillna(0).mean().to_dict()
    return avg_score


def normalize_score(score: Dict[int, float]) -> Dict[int, float]:
    """Normalize input score between min (=0) and max (=1)"""
    max_s = max(score.values())
    min_s = min(score.values())
    delta = max_s - min_s if max_s != min_s else 1
    return {key: (val - min_s) / delta for key, val in score.items()}


def average_ranks(ranks: Iterable[FeatureRanks]) -> FeatureRanks:
    n_feats = set(r.n_feats for r in ranks)
    if len(n_feats) > 1:
        raise ValueError("Input ranks refer to different features")
    n_feats = n_feats.pop()
    features = np.arange(n_feats)
    avg_ranks = []
    for f in features:
        avg_rank = np.average([rank[f] for rank in ranks])
        avg_ranks.append(avg_rank)
    return FeatureRanks(features=features, ranks=avg_ranks)


def get_best_n_features(ranks: FeatureRanks, n_to_keep: int) -> List[int]:
    ranks_data = ranks.get_data()
    sorted_data = sorted(ranks_data.items(), key=lambda x: x[1])
    feats = [feat for feat, _ in sorted_data[0:n_to_keep]]

    if len(feats) == n_to_keep:
        return feats

    # pad with non-present features, scramble to not introduce a bias
    all_feats = np.arange(ranks.n_feats)
    np.random.shuffle(all_feats)
    for f in all_feats:
        if f not in ranks_data:
            feats.append(f)
            if len(feats) == n_to_keep:
                return feats

    raise ValueError("Impossible to return so many best features")


def compute_t_student_p_value(sample: float, population: Iterable) -> float:
    """From Wikipeida:
    https://en.wikipedia.org/wiki/Prediction_interval#Unknown_mean,_unknown_variance
    Compute the p_value of the t-Student variable that represent the distribution
    of the n+1 sample of population, where `n = len(population)`.
    In few words, it gives the probability that `sample` belongs to `population`.
    The underlying assumption is that population is normally distributed
    with unknown mean and unknown variance

    Parameters
    ----------
        sample: float
            value for which we want the p-value
        population: Iterable
            values that represent the null hypothesis for `sample`

    Returns
    -------
        float
            the t-Student test p-value
    """
    m = np.mean(population)
    s = np.std(population)
    n = len(population)
    t = (sample - m) / (s * (1 + 1 / n) ** 0.5)
    return scipy.stats.t(df=n - 1).cdf(t)


def mute_loggers(loggers: List[str] = None):
    """Decorator to mute specific loggers within a function call. This is particularly
    useful when nested for loops might add excessive verbosity to a specific method.
    The input `loggers` is a list of loggers by name that should be muted. The logging
    level of the selected loggers is set to WARNING and restored to the original value
    at the end of the decorated method execution.

    Parameters
    ----------
    loggers : List[str], optional
        List of loggers by name that have to be muted, by default None
    """
    loggers = loggers or []

    def decorator(function: Callable):
        def decorated(*args, **kwargs):
            logs = {log_name: logging.getLogger(log_name) for log_name in loggers}
            old_levels = {
                log_name: log.getEffectiveLevel() for log_name, log in logs.items()
            }
            for log in logs.values():
                log.setLevel(logging.WARN)
            try:
                res = function(*args, **kwargs)
            finally:
                for log_name, log in logs.items():
                    log.setLevel(old_levels[log_name])
            return res

        return decorated

    return decorator



from typing import Callable, TypeVar, Union
from numpy import ndarray
from numpy.random import RandomState as RS
from sklearn.base import BaseEstimator


NumpyArray = ndarray
MetricFunction = Callable[[NumpyArray, NumpyArray], float]
GenericEstimator = TypeVar("GenericEstimator")
InputEstimator = Union[BaseEstimator, GenericEstimator]
RandomState = type(RS)



import pytest
import numpy as np
from py_muvr.models import sklearn_metrics


regression_true = np.array([1, 2, 3, 4, 5])
regression_pred = np.array([5, 4, 3, 2, 1])
classification_true = np.array([1, 0, 1, 2, 0, 0])
classification_pred = np.array([2, 0, 1, 1, 0, 1])
binary_classification_true = np.array([1, 0, 1, 1, 0, 0])
binary_classification_pred = np.array([1, 0, 1, 0, 1, 1])


@pytest.mark.parametrize(
    ("y_pred", "y_true", "res"),
    [
        (regression_pred, regression_true, -3),
        (regression_true, regression_true, 1),
    ],
)
def test_explained_variance(y_pred, y_true, res):
    metric = sklearn_metrics.SKLEARN_METRICS["explained_variance"]
    score = metric(y_true, y_pred)
    assert res == score


@pytest.mark.parametrize(
    ("y_pred", "y_true", "res"),
    [
        (regression_pred, regression_true, -3),
        (regression_true, regression_true, 1),
    ],
)
def test_r2(y_pred, y_true, res):
    metric = sklearn_metrics.SKLEARN_METRICS["r2"]
    score = metric(y_true, y_pred)
    assert res == score


@pytest.mark.parametrize(
    ("y_pred", "y_true", "res"),
    [
        (regression_pred, regression_true, -4),
        (regression_true, regression_true, 0),
    ],
)
def test_max_error(y_pred, y_true, res):
    metric = sklearn_metrics.SKLEARN_METRICS["max_error"]
    score = metric(y_true, y_pred)
    assert res == score


@pytest.mark.parametrize(
    ("y_pred", "y_true", "res"),
    [
        (regression_pred, regression_true, -2),
        (regression_true, regression_true, 0),
    ],
)
def test_neg_median_absolute_error(y_pred, y_true, res):
    metric = sklearn_metrics.SKLEARN_METRICS["neg_median_absolute_error"]
    score = metric(y_true, y_pred)
    assert res == score


@pytest.mark.parametrize(
    ("y_pred", "y_true", "res"),
    [
        (regression_pred, regression_true, -12 / 5),
        (regression_true, regression_true, 0),
    ],
)
def test_neg_mean_absolute_error(y_pred, y_true, res):
    metric = sklearn_metrics.SKLEARN_METRICS["neg_mean_absolute_error"]
    score = metric(y_true, y_pred)
    assert res == score


@pytest.mark.parametrize(
    ("y_pred", "y_true", "res"),
    [
        (regression_pred, regression_true, -8),
        (regression_true, regression_true, 0),
    ],
)
def test_neg_mean_squared_error(y_pred, y_true, res):
    metric = sklearn_metrics.SKLEARN_METRICS["neg_mean_squared_error"]
    score = metric(y_true, y_pred)
    assert res == score


@pytest.mark.parametrize(
    ("y_pred", "y_true", "res"),
    [
        (
            regression_pred,
            regression_true,
            -np.mean((np.log(1 + regression_true) - np.log(1 + regression_pred)) ** 2),
        ),
        (regression_true, regression_true, 0),
    ],
)
def test_neg_mean_squared_log_error(y_pred, y_true, res):
    metric = sklearn_metrics.SKLEARN_METRICS["neg_mean_squared_log_error"]
    score = metric(y_true, y_pred)
    assert res == score


@pytest.mark.parametrize(
    ("y_pred", "y_true", "res"),
    [
        (regression_pred, regression_true, -np.sqrt(8)),
        (regression_true, regression_true, 0),
    ],
)
def test_neg_root_mean_squared_error(y_pred, y_true, res):
    metric = sklearn_metrics.SKLEARN_METRICS["neg_root_mean_squared_error"]
    score = metric(y_true, y_pred)
    assert res == score


def test_neg_mean_poisson_deviance():
    metric = sklearn_metrics.SKLEARN_METRICS["neg_mean_poisson_deviance"]
    assert metric(regression_true, regression_pred) < 0
    assert metric(regression_true, regression_true) == 0


def test_neg_mean_gamma_deviance():
    metric = sklearn_metrics.SKLEARN_METRICS["neg_mean_gamma_deviance"]
    assert metric(regression_true, regression_pred) < 0
    assert metric(regression_true, regression_true) == 0


@pytest.mark.parametrize(
    ("y_pred", "y_true", "res"),
    [
        (classification_pred, classification_true, 0.5),
        (classification_true, classification_true, 1),
    ],
)
def test_accuracy(y_pred, y_true, res):
    metric = sklearn_metrics.SKLEARN_METRICS["accuracy"]
    score = metric(y_true, y_pred)
    assert res == score


@pytest.mark.parametrize(
    ("y_pred", "y_true", "res"),
    [
        (classification_pred, classification_true, 0.38888888888888884),
        (classification_true, classification_true, 1),
    ],
)
def test_balanced_accuracy(y_pred, y_true, res):
    metric = sklearn_metrics.SKLEARN_METRICS["balanced_accuracy"]
    score = metric(y_true, y_pred)
    assert res == score


@pytest.mark.parametrize(
    ("y_pred", "y_true", "res"),
    [
        (binary_classification_pred, binary_classification_true, 0.5),
        (binary_classification_true, binary_classification_true, 1),
    ],
)
def test_precision(y_pred, y_true, res):
    metric = sklearn_metrics.SKLEARN_METRICS["precision"]
    score = metric(y_true, y_pred)
    assert res == score


def test_precision_micro():
    metric = sklearn_metrics.SKLEARN_METRICS["precision_micro"]
    assert metric(classification_true, classification_pred) < 1
    assert metric(classification_true, classification_true) == 1


def test_precision_macro():
    metric = sklearn_metrics.SKLEARN_METRICS["precision_macro"]
    assert metric(classification_true, classification_pred) < 1
    assert metric(classification_true, classification_true) == 1


@pytest.mark.parametrize(
    ("y_pred", "y_true", "res"),
    [
        (binary_classification_pred, binary_classification_true, 2 / 3),
        (binary_classification_true, binary_classification_true, 1),
    ],
)
def test_recall(y_pred, y_true, res):
    metric = sklearn_metrics.SKLEARN_METRICS["recall"]
    score = metric(y_true, y_pred)
    assert res == score


def test_recall_macro():
    metric = sklearn_metrics.SKLEARN_METRICS["recall_macro"]
    assert metric(classification_true, classification_pred) < 1
    assert metric(classification_true, classification_true) == 1


def test_recall_micro():
    metric = sklearn_metrics.SKLEARN_METRICS["recall_micro"]
    assert metric(classification_true, classification_pred) < 1
    assert metric(classification_true, classification_true) == 1


@pytest.mark.parametrize(
    ("y_pred", "y_true", "res"),
    [
        (binary_classification_pred, binary_classification_true, 4 / 7),
        (binary_classification_true, binary_classification_true, 1),
    ],
)
def test_f1(y_pred, y_true, res):
    metric = sklearn_metrics.SKLEARN_METRICS["f1"]
    score = metric(y_true, y_pred)
    assert abs(res - score) < 1e-10


def test_f1_macro():
    metric = sklearn_metrics.SKLEARN_METRICS["f1_macro"]
    assert metric(classification_true, classification_pred) < 1
    assert metric(classification_true, classification_true) == 1


def test_f1_micro():
    metric = sklearn_metrics.SKLEARN_METRICS["f1_micro"]
    assert metric(classification_true, classification_pred) < 1
    assert metric(classification_true, classification_true) == 1


def test_jaccard():
    metric = sklearn_metrics.SKLEARN_METRICS["jaccard"]
    assert metric(binary_classification_true, binary_classification_pred) < 1
    assert metric(binary_classification_true, binary_classification_true) == 1


def test_jaccard_macro():
    metric = sklearn_metrics.SKLEARN_METRICS["jaccard_macro"]
    assert metric(classification_true, classification_pred) < 1
    assert metric(classification_true, classification_true) == 1


def test_jaccard_micro():
    metric = sklearn_metrics.SKLEARN_METRICS["jaccard_micro"]
    assert metric(classification_true, classification_pred) < 1
    assert metric(classification_true, classification_true) == 1



import pytest
from concurrent.futures import Future
from py_muvr.sync_executor import SyncExecutor


def square(x):
    return x ** 2


def bad_function(x):
    assert False


def test_sync_executor():
    executor = SyncExecutor()
    future = executor.submit(square, 2)
    assert isinstance(future, Future)
    assert future.result() == 4


def test_sync_executor_error():
    executor = SyncExecutor()
    future = executor.submit(bad_function, 2)
    assert isinstance(future, Future)
    with pytest.raises(AssertionError):
        future.result()
    isinstance(future.exception(), AssertionError)



from __future__ import annotations

from concurrent.futures._base import Future
from typing import Iterable, List, Dict, Union, Tuple
from dataclasses import dataclass
from py_muvr.data_structures import NumpyArray
from py_muvr.models.estimator import Estimator


@dataclass
class Split:
    id: int
    train_indices: NumpyArray
    test_indices: NumpyArray


@dataclass
class InputDataset:
    X: NumpyArray
    y: NumpyArray
    groups: NumpyArray

    @property
    def n_features(self):
        return self.X.shape[1]

    def __getitem__(self, data_slice: tuple) -> InputDataset:
        rows, features = data_slice
        sliced = InputDataset(
            X=self.X[rows, :], y=self.y[rows], groups=self.groups[rows]
        )
        if features:
            sliced.X = sliced.X[:, features]
        return sliced


@dataclass
class TrainTestData:
    train_data: InputDataset
    test_data: InputDataset


@dataclass
class FeatureRanks:
    def __init__(
        self, features: Iterable[int], ranks: Iterable[float], n_feats: int = None
    ):
        if n_feats is None:
            n_feats = max(features) + 1

        if max(features) >= n_feats:
            raise ValueError("n_feats should be greater than max(features)")
        if max(ranks) > n_feats:
            raise ValueError("ranks should range from 1 to n_feats at most")
        self.n_feats = n_feats
        self._data = dict(zip(features, ranks))

    def __getitem__(self, feature: int):
        if feature >= self.n_feats:
            raise ValueError(
                f"The feature ranks are relative to {self.n_feats} features at most"
            )
        return self._data.get(feature, self.n_feats)

    def get_data(self):
        return self._data.copy()


@dataclass
class FeatureEvaluationResults:
    ranks: FeatureRanks
    test_score: float
    model: Estimator


InnerLoopResults = List[FeatureEvaluationResults]


@dataclass
class SelectedFeatures:
    min: Union[List[int], NumpyArray, List[str], Tuple[int]]
    max: Union[List[int], NumpyArray, List[str], Tuple[int]]
    mid: Union[List[int], NumpyArray, List[str], Tuple[int]]

    def __getitem__(self, item):
        accepted_keys = {"min", "mid", "max"}
        if item in accepted_keys:
            return getattr(self, item)
        else:
            raise KeyError(f"Accepted keys are: {accepted_keys}")


@dataclass
class FeatureEliminationResults:
    n_features_to_score_map: Dict[int, float]
    best_features: SelectedFeatures


@dataclass
class OuterLoopResults:
    min_eval: FeatureEvaluationResults
    max_eval: FeatureEvaluationResults
    mid_eval: FeatureEvaluationResults
    n_features_to_score_map: Dict[int, float]


@dataclass
class ScoreCurve:
    n_features: List[int]
    scores: List[float]


FeatureSelectionRawResults = List[List[Union[OuterLoopResults, Future]]]


@dataclass
class FeatureSelectionResults:
    raw_results: FeatureSelectionRawResults
    selected_features: SelectedFeatures
    score_curves: Dict[str, List[ScoreCurve]]
    selected_feature_names: SelectedFeatures = None

    def __repr__(self):
        return "FeatureSelectionResults"



import pytest
import numpy as np
from py_muvr.models.sklearn_metrics import SKLEARN_METRICS
from py_muvr.models.metrics import make_metric, _make_metric_from_string, miss_score


def test_make_metric():
    metric = make_metric("MISS")
    assert metric
    assert metric is miss_score


def test_miss_score():
    y_pred = np.array([1, 0, 0, 1])
    y_true = np.array([0, 1, 1, 0])
    assert miss_score(y_true, y_pred) == -4
    y_true = np.array([1, 0, 0, 1])
    assert miss_score(y_true, y_pred) == 0
    y_true = np.array([1, 0, 1, 0])
    assert miss_score(y_true, y_pred) == -2


def test_make_metric_from_string():
    for metric_id in SKLEARN_METRICS:
        assert _make_metric_from_string(metric_id)
    assert _make_metric_from_string("MISS") is miss_score
    with pytest.raises(ValueError):
        _make_metric_from_string("yo")



import logging
from typing import List, Union, Iterable
from matplotlib.pyplot import Figure
import matplotlib.ticker as mtick
import pandas as pd
from py_muvr.permutation_test import PermutationTest
from matplotlib import pyplot as plt

from py_muvr.data_structures import FeatureSelectionResults

log = logging.getLogger(__name__)


class PALETTE:
    lightblue = "#deebf7"
    blue = "#3182bd"
    black = "black"
    white = "white"
    grey = "grey"
    lightgrey = "#9facbd"


def plot_validation_curves(
    feature_selection_results: FeatureSelectionResults, **figure_kwargs
) -> plt.Figure:
    curves = feature_selection_results.score_curves
    plt.figure(**figure_kwargs)
    for i, curve in enumerate(curves["outer_loops"]):
        label = "Outer loop average" if i == 0 else None
        plt.semilogx(curve.n_features, curve.scores, c=PALETTE.lightblue, label=label)
    for i, curve in enumerate(curves["repetitions"]):
        label = "Repetition average" if i == 0 else None
        plt.semilogx(curve.n_features, curve.scores, c=PALETTE.blue, label=label)
    for i, curve in enumerate(curves["total"]):
        label = "Total average" if i == 0 else None
        plt.semilogx(curve.n_features, curve.scores, c=PALETTE.black, label=label)

    min_y, max_y = plt.gca().get_ylim()
    selected_features = feature_selection_results.selected_features
    for attribute in ["min", "max", "mid"]:
        n_feats = len(getattr(selected_features, attribute))
        plt.vlines(
            n_feats,
            min_y,
            max_y,
            linestyle="--",
            colors=PALETTE.grey,
            lw=2,
            label=attribute,
            zorder=100000,
        )

    plt.xlabel("# features")
    plt.ylabel("Fitness score")
    plt.grid(ls=":")
    plt.legend(bbox_to_anchor=(1.05, 1), loc="upper left", borderaxespad=0)
    return plt.gcf()


def plot_feature_rank(
    feature_selection_results: FeatureSelectionResults,
    model: str,
    feature_names: List[str] = None,
    show_outliers: bool = True,
    **figure_kwargs,
) -> Figure:
    if model not in {"min", "max", "mid"}:
        raise ValueError("The model parameter must be one of 'min', 'max' or 'mid'.")

    eval_attr = model + "_eval"
    feats_attr = model

    ranks = []
    for r in feature_selection_results.raw_results:
        for ol in r:
            ranks_raw_data = getattr(ol, eval_attr).ranks.get_data()
            ranks.append(ranks_raw_data)

    selected_features = feature_selection_results.selected_features
    best = getattr(selected_features, feats_attr)
    selected_ranks = pd.DataFrame(r for r in ranks)[best]

    sorted_feats = selected_ranks.mean().sort_values().index
    selected_ranks = selected_ranks[sorted_feats]

    if "figsize" not in figure_kwargs.keys():
        fig_width = len(selected_ranks.columns) / 3
        figure_kwargs["figsize"] = (6, max(fig_width, 5))

    fig, (ax_ranks, ax_notnan) = plt.subplots(
        nrows=1, ncols=2, sharey=True, **figure_kwargs
    )

    ax_notnan.xaxis.set_major_formatter(mtick.PercentFormatter())
    ax_notnan.set_ylabel("Feature")
    ax_notnan.set_xlabel("Percentage of times selected")
    ax_ranks.set_xlabel("Feature Rank")

    for ax in [ax_notnan, ax_ranks]:
        ax.grid(linestyle=":", zorder=0)
        ax.tick_params(axis="x")
        ax.xaxis.tick_top()
        ax.xaxis.set_label_position("top")

    bbox_props = {
        "color": PALETTE.blue,
        "alpha": 0.8,
    }
    bbox_color = {"boxes": PALETTE.blue, "medians": PALETTE.black}

    if feature_names is not None:
        feature_numbers = range(len(feature_names))
        numbers_to_names = dict(zip(feature_numbers, feature_names))
        selected_ranks.rename(columns=numbers_to_names, inplace=True)

    selected_ranks.boxplot(
        positions=range(len(selected_ranks.columns)),
        color=bbox_color,
        patch_artist=True,
        ax=ax_ranks,
        boxprops=bbox_props,
        vert=False,
        showfliers=show_outliers,
    )

    (selected_ranks.notna().mean() * 100).plot.barh(
        facecolor=PALETTE.lightgrey,
        ax=ax_notnan,
        edgecolor=PALETTE.black,
        grid=True,
        alpha=0.8,
    )

    ax_notnan.invert_yaxis()  # being the y-axis shared, it will invert both

    fig.tight_layout()  # otherwise the right y-label is slightly clipped

    return fig


def plot_permutation_scores(
    permutation_test: PermutationTest,
    model: str,
    bins: Union[int, str, Iterable[float]] = "auto",
    **fig_kwargs,
) -> Figure:
    score, perm_scores = permutation_test.compute_permutation_scores(model)
    p_value = permutation_test.compute_p_values(model, ranks=False)
    fig, ax = plt.subplots(1, 1, **fig_kwargs)
    ax.grid(linestyle=":", zorder=0)
    counts, _, _ = ax.hist(
        perm_scores,
        bins=bins,
        alpha=0.8,
        edgecolor=PALETTE.white,
        facecolor=PALETTE.blue,
        label="Permutation Scores",
        zorder=10,
    )
    ax.vlines(
        score,
        ymin=0,
        ymax=counts.max(),
        color=PALETTE.black,
        label="Feature Selection Score",
        zorder=20,
    )
    ax.set_ylabel("Number of Occurrences")
    ax.set_xlabel("Score")
    ax.legend(bbox_to_anchor=(1.05, 1), loc="upper left", borderaxespad=0)
    ax.set_title("Feature selection p-value = %1.3g" % p_value)
    return fig



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""
import versioneer
from setuptools import setup, find_packages

with open("docs/readme.rst") as readme_file:
    readme = readme_file.read()

with open("HISTORY.rst") as history_file:
    history = history_file.read()

requirements = []

setup_requirements = [
    "pytest-runner",
]

test_requirements = [
    "pytest",
]

setup(
    author="Data Revenue GmbH",
    author_email="carlos@datarevenue.com",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
    description="Nested cross validation for feature selection in Python",
    install_requires=[
        "matplotlib>=3.3.2",
        "numpy>=1.19.2",
        "pandas>=1.1.3",
        "scipy>=1.5.2",
        "scikit-learn>=0.23.2",
        "xgboost>=1.3.1",
        "progressbar2>=3.53.1",
    ],
    license="MIT license",
    long_description=readme + "\n\n" + history,
    include_package_data=True,
    keywords="muvr, muv",
    name="py-muvr",
    packages=find_packages(exclude=["tests"]),
    setup_requires=setup_requirements,
    test_suite="tests",
    tests_require=test_requirements,
    url="https://github.com/datarevenue-berlin/py-MUVR",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    zip_safe=False,
)



import numpy as np
from sklearn import clone
from sklearn.base import BaseEstimator

from py_muvr.data_structures import RandomState
from py_muvr.models.estimator import Estimator
from py_muvr.data_structures.data_types import NumpyArray


class ScikitLearnEstimator(Estimator):
    """Wraps scikit-learn models. This class exposes the fit` and `predict` methods of
    its input estimator. The estimator must be a scikit-learn model, such as
    RandomForestClassifier, LogisticRegression, etc.
    Additionally, it can clone itself with the method `clone`. This is useful
    if the class should remain state-less.
    The input random state is held for cloning purposes and it's used to override the
    random state of the input estimator.
    The `feature_importance` exposed by this class depends on the input estimator:
    If the estimator has a `feature_importance_` attribute, than it is returned,
    otherwise the absolute values of the `coef_[0]` vector are returned. This means
    that in the exotic case of an estimator implementing both, `feature_importance_` is
    returned

    Parameters
    ----------
    estimator : BaseEstimator
        The wrapped estimator
    random_state: RandomState
        a random state instance to control reproducibility
    """

    def __init__(self, estimator: BaseEstimator, random_state: RandomState):
        self._estimator = estimator
        self._random_state = random_state
        self.set_random_state(random_state)

    @property
    def _estimator_type(self):
        return self._estimator._estimator_type

    @property
    def feature_importances(self):
        feature_importances = self._get_feature_importances(self._estimator)
        if feature_importances is None:
            raise ValueError(
                f"The estimator provided {self._estimator.__repr__()} "
                f"has no feature importances"
            )
        return feature_importances

    def set_random_state(self, random_state: RandomState):
        try:
            self._estimator.set_params(random_state=random_state)
        except ValueError:
            pass  # not all models have a random_state param (e.g. LinearRegression)

    def clone(self):
        estimator_clone = clone(self._estimator)
        return self.__class__(estimator_clone, self._random_state)

    def fit(self, X, y):
        self._estimator.fit(X, y)
        return self

    def predict(self, X):
        return self._estimator.predict(X)

    def __repr__(self):
        return f"SKLearnEstimator(model={self._estimator.__repr__()})"

    @staticmethod
    def _get_feature_importances(estimator: BaseEstimator) -> NumpyArray:
        if hasattr(estimator, "feature_importances_"):
            return estimator.feature_importances_
        if hasattr(estimator, "coef_"):
            coefficients = np.abs(estimator.coef_)
            if coefficients.ndim == 1:
                return coefficients
            else:
                return coefficients[0]
        return None


class ScikitLearnPipeline(ScikitLearnEstimator):
    """Extends ScikitLearnEstimator to scikitlearn pipelines. The  feature importance
    of the pipeline is the first attribute found between `feature_importances_` or
    `abs(coef_[0])` among its various steps. This should normally correspond to the
    last one, but more unorthodox cases are supported.
    """

    @property
    def feature_importances(self):
        for _, step in self._estimator.steps:
            feature_importances = self._get_feature_importances(step)
            if feature_importances is not None:
                return feature_importances
        raise ValueError(
            f"The estimator provided {self._estimator.__repr__()} "
            f"has no feature importances"
        )

    def set_random_state(self, random_state: RandomState):
        for _, step in self._estimator.steps:
            try:
                step.set_params(random_state=random_state)
            except ValueError:
                pass  # not all the elements of the pipeline have to be random

    def __repr__(self):
        return f"SKLearnPipeline(n_steps={len(self._estimator.steps)})"



import pytest
import numpy as np
from py_muvr.data_structures import InputDataset, Split
from py_muvr.data_splitter import DataSplitter


@pytest.fixture
def grouped_dataset():
    X = np.zeros((10, 10))
    y = np.arange(10)
    groups = y // 2
    return InputDataset(X=X, y=y, groups=groups)


def test_data_splitter(dataset):
    ds = DataSplitter(n_outer=5, n_inner=4, random_state=0, input_data=dataset)
    assert ds
    assert ds._splits


def test_make_splits(dataset):
    ds = DataSplitter(n_outer=5, n_inner=4, random_state=0, input_data=dataset)
    assert ds._splits
    assert len(ds._splits) == 5 * 4 + 5
    split = ds._splits[(0, 2)]
    train_idx = split.train_indices
    test_idx = split.test_indices
    assert len(train_idx) == 6
    assert len(test_idx) == 3
    split = ds._splits[(0, None)]
    train_idx = split.train_indices
    test_idx = split.test_indices
    assert len(train_idx) == 9
    assert len(test_idx) == 3
    assert not set(train_idx).intersection(test_idx)


def test_split_separation(dataset):
    ds = DataSplitter(n_outer=5, n_inner=4, random_state=0, input_data=dataset)
    # all indices should appear once if joining outer_test, inner_test and inner_train
    for outer_split in ds.iter_outer_splits():
        for inner_split in ds.iter_inner_splits(outer_split):
            all_indeces = (
                list(outer_split.test_indices)
                + list(inner_split.test_indices)
                + list(inner_split.train_indices)
            )
            assert len(all_indeces) == 12
            assert sorted(all_indeces) == list(range(12))
            out_train = list(inner_split.test_indices) + list(inner_split.train_indices)
            assert sorted(out_train) == sorted(outer_split.train_indices)


def test_make_splits_grouped(grouped_dataset):
    ds = DataSplitter(n_outer=5, n_inner=4, random_state=0, input_data=grouped_dataset)
    groups = grouped_dataset.groups
    assert ds
    # check there is no intersection among the groups
    for outer_split in ds.iter_outer_splits():
        train_idx = outer_split.train_indices
        test_idx = outer_split.test_indices
        for inner_split in ds.iter_inner_splits(outer_split):
            inner_train = inner_split.train_indices
            valid_idx = inner_split.test_indices
            assert not set(groups[inner_train]).intersection(groups[valid_idx])
            assert not set(groups[test_idx]).intersection(groups[valid_idx])
            assert not set(groups[inner_train]).intersection(groups[test_idx])


def test_iter_outer_splits(dataset):
    ds = DataSplitter(n_outer=5, n_inner=4, random_state=0, input_data=dataset)
    for outer_split in ds.iter_outer_splits():
        assert outer_split
        assert outer_split == ds._splits[(outer_split.id, None)]


def test_iter_inner_splits(dataset):
    ds = DataSplitter(n_outer=5, n_inner=4, random_state=0, input_data=dataset)
    for outer_split in ds.iter_outer_splits():
        for inner_split in ds.iter_inner_splits(outer_split):
            assert inner_split
            assert inner_split == ds._splits[(outer_split.id, inner_split.id)]


def test_randomness(dataset):
    random_state = np.random.RandomState(0)
    ds = DataSplitter(
        n_outer=5, n_inner=4, random_state=random_state, input_data=dataset
    )
    ds2 = DataSplitter(
        n_outer=5, n_inner=4, random_state=random_state, input_data=dataset
    )
    assert all(
        sorted(ds._splits[k].train_indices) != sorted(ds2._splits[k].train_indices)
        for k in ds._splits
    )


def test_non_randomness(dataset):
    ds = DataSplitter(n_outer=5, n_inner=4, random_state=0, input_data=dataset)
    ds2 = DataSplitter(n_outer=5, n_inner=4, random_state=0, input_data=dataset)
    assert all(
        sorted(ds._splits[k].train_indices) == sorted(ds2._splits[k].train_indices)
        for k in ds._splits
    )


def test_split_data(dataset):
    ds = DataSplitter(n_outer=5, n_inner=4, random_state=0, input_data=dataset)
    split = Split(0, [0, 1, 2], [3, 4, 5, 6])
    split_data = ds.split_data(dataset, split, features=[0, 1, 2])
    assert split_data.test_data.X.shape == (4, 3)
    assert split_data.test_data.y.size == 4
    assert split_data.train_data.X.shape == (3, 3)
    assert split_data.train_data.y.size == 3



# -*- coding: utf-8 -*-

"""Top-level package for py_muvr."""

__author__ = """Data Revenue GmbH"""
__email__ = "giulio@datarevenue.com"
__version__ = "0.1.0"

import logging
import sys
from py_muvr.feature_selector import FeatureSelector
from ._version import get_versions


__version__ = get_versions()["version"]
del get_versions

logger = logging.getLogger("py_muvr")
logger.setLevel(level=logging.INFO)
formatter = logging.Formatter(
    "%(asctime)s %(name)-26s %(levelname)-7s %(message)s", datefmt="%Y-%m-%d %H:%M:%S"
)
ch = logging.StreamHandler(sys.stdout)
ch.setFormatter(formatter)
logger.addHandler(ch)



from py_muvr.data_structures.data_types import (
    NumpyArray,
    MetricFunction,
    InputEstimator,
    RandomState,
)
from py_muvr.data_structures.data_models import (
    Split,
    InputDataset,
    TrainTestData,
    FeatureRanks,
    FeatureEvaluationResults,
    InnerLoopResults,
    SelectedFeatures,
    FeatureEliminationResults,
    OuterLoopResults,
    ScoreCurve,
    FeatureSelectionRawResults,
    FeatureSelectionResults,
)
